#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2020 Synology Inc. All rights reserved.

import sys, os, signal, json, six
from hypervisor import defines

import logging
import logging.handlers

logger = logging.getLogger(defines.VM_LOGGER_NAME)
logger.setLevel(logging.DEBUG)

handler = logging.handlers.SysLogHandler(address='/dev/log')
handler.setLevel(logging.INFO)

formatter = logging.Formatter('[%(levelname)s] synoabk_vm::%(filename)s:%(lineno)d(%(process)d,%(thread)d) %(message)s')
handler.setFormatter(formatter)

logger.addHandler(handler)

# run daemon
if __name__ == '__main__':
    if six.PY3:
        from importlib import reload
        reload(sys)
    elif six.PY2:
        reload(sys)
        sys.setdefaultencoding('utf-8')

    def HandleRequest(request, progressFilePath):
        command = request['command']
        argument = request['argument']
        inventory = request['inventory']
        server = GetServer(inventory, progressFilePath, argument['log_result_id'])
        argument.pop('log_result_id')
        logger.info(command)
        logger.info({k:v for k,v in argument.items() if k not in ['vm_user', 'vm_pwd']} if argument else None)
        logger.info('progressFilePath: ' + progressFilePath)
        return server.dispatch(command, argument)

    def GetServer(inventory, progressFilePath = None, logResultId = 0):
        try:
            inventoryId = inventory['inventory_id']
            hosttype    = inventory['host_type']
            hostaddr    = inventory['host_addr']
            username    = inventory['auth_user']
            password    = inventory['auth_password']
            thumbprint  = inventory['auth_thumbprint']
            port        = inventory['port_webapi']
            protocol    = inventory['protocol']

            oldInventory, server, taskPool = None, None, {}

            if  hosttype == defines.HostType.ESXi.value or \
                hosttype == defines.HostType.vCenter.value:
                from hypervisor import vsphere_server
                server = vsphere_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool, progressFilePath, logResultId)
            elif hosttype == defines.HostType.HyperV.value or \
                    hosttype == defines.HostType.FailoverCluster.value:
                from hypervisor import hyperv_server
                server = hyperv_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool, hosttype, progressFilePath, logResultId)
            elif hosttype == defines.HostType.SCVMM.value:
                from hypervisor import scvmm_server
                server = scvmm_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool)
            else:
                raise Exception('Unknown Host Type %d' % hosttype)
        except Exception as e:
                import traceback
                logger.error(traceback.format_exc())
        finally:
            pass
        return server

    def HandleSignal(sig, frame):
        if sig == signal.SIGPIPE:
            return
        if sig == signal.SIGTERM or sig == signal.SIGINT:
            os.environ["IS_PRIVILEGE_WORKER_CANCELED"] = "1"
            return
        os._exit(0)

    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGPIPE):
        signal.signal(sig, HandleSignal)

    StrRequest = os.getenv('REQUEST')
    Request = json.loads(StrRequest)
    ProgressFilePath = sys.argv[1]
    HandleRequest(Request, ProgressFilePath)
