# coding=utf8

from reportlab.lib.colors import HexColor, toColor
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.shapes import Drawing, _DrawingEditorMixin
from reportlab.lib.validators import Auto
from reportlab.graphics.charts.barcharts import HorizontalBarChart
from reportlab.graphics.charts.textlabels import Label

from util import StringManager, get_unit
from render import render_size
from data_util import config_map

class HBarChart(_DrawingEditorMixin,Drawing):
	def __init__(self, seriesList, width=1188, height=220, scale=2.0, *args, **kw):
		Drawing.__init__(self, width * scale, height * scale, *args, **kw)

		fontName      = StringManager.get_font_name()
		legendHExpand = 0 * scale
		legenVExpand  = 26 * scale
		topPadding    = 10 * scale
		leftPadding   = 0 * scale
		rightPadding  = 0 * scale
		bottomPadding = 0 * scale
		# chart anchor on nw rather than original sw
		chartLeft = leftPadding
		chartTop = self.height - topPadding - legenVExpand

		if not seriesList:
			max_value = 0.0
			empty_series = {
				'values': [0]
			}
			empty_series.update(config_map[''])
			seriesList = [empty_series]
		else:
			max_value = max(list(
				map(lambda group: sum(group), list(
					zip(*[series['values'] for series in seriesList])
				))
			))
		unit_str, unit_value = get_unit(max_value)

		#chart
		self._add(self, HorizontalBarChart(), name='chart', validate=None, desc=None)
		self.chart.x      = chartLeft
		self.chart.y      = bottomPadding
		self.chart.width  = self.width - self.chart.x - rightPadding - legendHExpand
		self.chart.height = chartTop   - self.chart.y
		self.chart.fillColor           = None
		self.chart.strokeColor         = None
		self.chart.strokeWidth         = 1
		self.chart.valueAxis.forceZero = 1

		# bars
		self.chart.bars.strokeWidth    = 0
		self.chart.bars.strokeColor    = None
		self.chart.barLabelFormat      = lambda v: render_size(v) if max_value != 0 and float(v) / max_value >= 0.01 else None
		self.chart.barLabels.boxTarget = 'mid'
		self.chart.barLabels.boxAnchor = 'c'
		self.chart.barLabels.fillColor = HexColor('#FFFFFF')
		self.chart.barLabels.fontName  = fontName
		self.chart.barLabels.fontSize  = 12 * scale
		self.chart.useAbsolute         = 3 # barWidth & groupSpacing is absolute
		self.chart.barWidth            = 14 * scale
		self.chart.groupSpacing        = 0 * scale

		# x-axis
		self.chart.categoryAxis.labels.fontName  = fontName
		self.chart.categoryAxis.labels.fontSize  = 12 * scale
		self.chart.categoryAxis.strokeWidth      = 0
		self.chart.categoryAxis.strokeColor      = None
		self.chart.categoryAxis.style            = 'mixed'
		self.chart.categoryAxis.visibleTicks     = 0
		self.chart.categoryAxis.visibleAxis      = 0

		# y-axis
		self.chart.valueAxis.labels.fontName     = fontName
		self.chart.valueAxis.labels.fontSize     = 12 * scale
		self.chart.valueAxis.labelTextFormat     = '%0.2f '
		self.chart.valueAxis.strokeWidth         = 0
		self.chart.valueAxis.gridStrokeColor     = toColor('rgba(138,146,153,0.15)')
		self.chart.valueAxis.gridStrokeWidth     = 1
		self.chart.valueAxis.visibleTicks        = 0
		self.chart.valueAxis.visibleAxis         = 0
		self.chart.valueAxis.visibleGrid         = 0
		self.chart.valueAxis.avoidBoundFrac      = 0

		# legend
		self._add(self, Legend(), name='legend', validate=None, desc=None)
		self.legend.alignment        = 'right'
		self.legend.boxAnchor        = 'sw'
		self.legend.x                = chartLeft + 2 * scale
		self.legend.y                = chartTop + 9 * scale
		self.legend.dx               = 12 * scale
		self.legend.dy               = 12 * scale
		self.legend.swdx             = 0 * scale
		self.legend.swdy             = 0 * scale
		self.legend.columnMaximum    = 1
		self.legend.deltax           = 130 * scale
		self.legend.deltay           = 0 * scale
		self.legend.dxTextSpace      = 5 * scale
		self.legend.fontName         = fontName
		self.legend.fontSize         = 12 * scale
		self.legend.strokeColor      = None
		self.legend.strokeWidth      = 0
		self.legend.colorNamePairs   = [(HexColor(series['color']), StringManager.get(series['display_short_text'])) for series in seriesList]

		for i, series in enumerate(seriesList):
			self.chart.bars[i].name = series['label']
			self.chart.bars[i].fillColor = HexColor(series['color'])

		categoryCount = len(seriesList)
		self.chart.seriesOrder = (tuple(i for i in range(max(1, categoryCount))),)
		self.chart.data = [series['values'] for series in seriesList]


class SeparateHBarChart(_DrawingEditorMixin,Drawing):
	def __init__(self, seriesList, width=1188, height=220, scale=2.0, *args, **kw):
		Drawing.__init__(self, width * scale, height * scale, *args, **kw)

		fontName      = StringManager.get_font_name()
		legendHExpand = 0 * scale
		legenVExpand  = 0 * scale
		topPadding    = 0 * scale
		leftPadding   = 26 * scale
		rightPadding  = 26 * scale
		bottomPadding = -10 * scale
		# chart anchor on nw rather than original sw
		chartLeft = leftPadding
		chartTop  = self.height - topPadding - legenVExpand

		self._add(self, HorizontalBarChart(), name='chart', validate=None, desc=None)

		#chart
		self.chart.x      = chartLeft
		self.chart.y      = bottomPadding
		self.chart.width  = self.width - self.chart.x - rightPadding - legendHExpand
		self.chart.height = chartTop   - self.chart.y
		self.chart.fillColor           = None
		self.chart.strokeColor         = None
		self.chart.strokeWidth         = 1
		self.chart.valueAxis.forceZero = 1

		# bars
		self.chart.bars.strokeWidth   = 0
		self.chart.bars.strokeColor   = None
		self.chart.useAbsolute        = 7 # barWidth & groupSpacing & barSpacing is absolute
		self.chart.barWidth           = 7 * scale
		self.chart.groupSpacing       = 56 * scale - self.chart.barWidth

		# x-axis
		self.chart.categoryAxis.labels.fontName  = fontName
		self.chart.categoryAxis.labels.fontSize  = 12 * scale
		self.chart.categoryAxis.strokeWidth      = 0
		self.chart.categoryAxis.strokeColor      = None
		self.chart.categoryAxis.style            = 'stacked'
		self.chart.categoryAxis.visibleTicks     = 0
		self.chart.categoryAxis.visibleAxis      = 0

		# y-axis
		self.chart.valueAxis.labels.fontName = fontName
		self.chart.valueAxis.labels.fontSize = 12 * scale
		self.chart.valueAxis.labelTextFormat = '%0.1f'
		self.chart.valueAxis.strokeWidth     = 0
		self.chart.valueAxis.gridStrokeColor = toColor('rgba(138,146,153,0.15)')
		self.chart.valueAxis.gridStrokeWidth = 1
		self.chart.valueAxis.visibleTicks    = 0
		self.chart.valueAxis.visibleAxis     = 0
		self.chart.valueAxis.visibleGrid     = 0
		self.chart.valueAxis.visibleLabels   = 0
		self.chart.valueAxis.avoidBoundFrac  = 0

		self.chart.bars[0].name = 'count'
		self.chart.bars[0].fillColor = HexColor('#3876BE')
		self.chart.bars[1].name = 'rest'
		self.chart.bars[1].fillColor = HexColor('#DBE0E4')

		categoryCount = len(seriesList)
		reversedSeriesList = reversed(seriesList)
		data = [series['values'][0] for series in reversedSeriesList]
		maxValue = max(data)

		self.chart.data = [data, [maxValue - d for d in data]]

		for idx, series in enumerate(seriesList):
			# device type
			name = 'type_' + series['label']
			self._add(self, Label(), name=name, validate=None, desc=None)
			typeLabel = getattr(self, name)
			# device count
			name = 'type_' + series['label']
			self._add(self, Label(), name=name, validate=None, desc=None)
			countLabel = getattr(self, name)

			typeLabel.y  = countLabel.y  = chartTop - (0.5 + idx) * 56 * scale
			typeLabel.dy = countLabel.dy = 4 * scale
			typeLabel.fontName = countLabel.fontName = fontName
			typeLabel.fontSize = countLabel.fontSize = 12 * scale

			typeLabel.boxAnchor = 'sw'
			typeLabel.x         = chartLeft
			typeLabel._text     = StringManager.get(series['display_text'])

			countLabel.boxAnchor = 'se'
			countLabel.x         = chartLeft + self.chart.width
			countLabel._text     = StringManager.get('report:device_count', series['values'][0])
