# -*- coding: utf-8 -*-

import jinja2, js2py
from datetime import datetime, timedelta
from math import ceil
import json, re, cgi, math, calendar, locale
from functools import reduce
from six import text_type, PY2

if PY2:
	from HTMLParser import HTMLParser
	from cgi import escape
else:
	from html.parser import HTMLParser
	from html import escape

output_format = 'pdf'

def get_str(string_path, *args, **kwargs):
	if output_format == 'pdf':
		ret = StringManager.get(string_path, *args, **kwargs)

		# NOTE: uncomment this to support CJK / other mixed string
		# if StringManager.is_support_fonts() and not kwargs.get('no_font', False):
		#   parser = StringManager.DecoratedStringParser()
		#   parser.feed(ret)
		#   ret = parser.decorated

		return ret
	elif output_format == 'html':
		if args:
			return ''
		else:
			return '<{1} class="LABSS">{0}</{1}>'.format(string_path, kwargs.get('tag', 'span'))
	else:
		return string_path

def get_rich_str(str_data):
	if type(str_data) == tuple:
		if len(str_data) == 2 and type(str_data[1]) == dict:
			return get_str(str_data[0], **str_data[1])
		else:
			return get_str(*str_data)
	else:
		return str_data

def load_js_as_json(path):
	content = ''
	first_line = True
	with open(path, 'r') as f:
		for line in f:
			line = line.strip()
			if first_line:
				first_line = False
				start = line.find('{')
				if start >= 0:
					line = line[start:]
			if line.endswith(';'):
				content = content + line.rstrip(';')
				break
			else:
				content = content + line
	return json.loads(content)

def find_text_break_point(text, font_size, box_width, ellipsis='', max_line=-1):
	accum_width = StringManager.get_text_width(ellipsis + text, font_size)
	ellipsis_len = len(ellipsis)
	need_ellipsis = False

	if accum_width[-1] - accum_width[ellipsis_len] <= box_width:
		return [], need_ellipsis

	# find all line-break-point
	break_point = []
	if max_line <= 0:
		# assume container will expand for lines
		line_start = 0
		for line_end in range(line_start + 1, len(accum_width)):
			if (accum_width[line_end] - accum_width[line_start]) > box_width:
				break_point.append(line_end - 1)
				line_start = line_end - 1
	else:
		if ellipsis:
			line_start = ellipsis_len
			for line_end in range(line_start + 1, len(accum_width)):
				if (accum_width[line_end] - accum_width[line_start]) > box_width:
					if len(break_point) + 1 >= max_line:
						need_ellipsis = True
						for idx in range(line_end, line_start, -1):
							if (accum_width[idx] - accum_width[line_start]) + accum_width[ellipsis_len] <= box_width:
								break_point.append(idx - 1 - ellipsis_len)
								break
						break
					else:
						break_point.append(line_end - 1 - ellipsis_len)
						line_start = line_end - 1
		else:
			# assume font size will shrink to fit container
			line_start = 0
			for line_end in range(line_start + 1, len(accum_width)):
				if (accum_width[line_end] - accum_width[line_start]) > box_width:
					break_point.append(line_end - 1)
					line_start = line_end - 1

			origin_line_count = len(break_point) + 1
			if origin_line_count > max_line:
				line_count = origin_line_count
				for line_count in range(max_line + 1, origin_line_count):
					magic_factor = 1.08 * float(max_line) / line_count
					break_point = []
					line_start = 0
					for line_end in range(line_start + 1, len(accum_width)):
						if (accum_width[line_end] - accum_width[line_start]) * magic_factor > box_width:
							break_point.append(line_end - 1)
							line_start = line_end - 1
					if len(break_point) + 1 <= line_count:
						break
				break_point += [len(text)] * (line_count - len(break_point) - 1)

	return break_point, need_ellipsis

def break_text(text, font_size, box_width, ellipsis='', max_line=-1):
	break_point, need_ellipsis = find_text_break_point(text, font_size, box_width, ellipsis=ellipsis, max_line=max_line)

	break_point = [0] + break_point
	lines = list(text[break_point[idx -1]:break_point[idx]] for idx in range(1, len(break_point)))
	if need_ellipsis:
		lines[-1] += ellipsis
	else:
		lines.append(text[break_point[-1]:])

	return lines

class StringManager:
	_table = {}
	month_str = []
	lang = 'enu'
	_log_ctx = None
	_major_font = None
	_minor_font_list = []
	_week_str = []
	_month_str = []

	class DecoratedStringParser(HTMLParser):
		def __init__(self):
			HTMLParser.__init__(self)
			self.decorated = ''

		def handle_starttag(self, tag, attrs):
			self.decorated += '<{}{}>'.format(tag, ''.join([' {}="{}"'.format(attr[0],attr[1]) for attr in attrs]))

		def handle_endtag(self, tag):
			self.decorated += '</{}>'.format(tag)

		def handle_data(self, data):
			self.decorated += StringManager.decorate_text(data)

	@classmethod
	def initialize(cls, lang, string_path, script_path):
		cls.lang = lang
		cls._table = load_js_as_json(string_path)

		cls._log_ctx = js2py.EvalJs()
		cls._log_ctx.execute('''
			var window = this;
			function _LABS(sec, key) { return lang_table[sec][key] || ''; };
			var lang_table =''' + json.dumps(cls._table[cls.lang]) + ';')
		for js_file in (script_path,):
			with open(js_file, 'r') as f:
				cls._log_ctx.execute(f.read())

		cls._week_str = [cls.get('schedule:' + key) for key in ('monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday')]
		cls._month_str = [u''] + [cls.get('schedule:month_abbr' + str(i + 1)) for i in range(12)]

	@classmethod
	def get_month_name(cls, idx):
		return cls._month_str[idx]

	@classmethod
	def get_week_name(cls, idx):
		return cls._week_str[idx]

	@classmethod
	def register_font(cls, fontName, filename, charWidths, defaultWidth):
		if cls._major_font is None:
			cls._major_font = (fontName, filename, charWidths, defaultWidth)
		else:
			cls._minor_font_list.append((fontName, filename, charWidths, defaultWidth))

	@classmethod
	def is_support_fonts(cls):
		return bool(cls._minor_font_list)

	@classmethod
	def get_font_name(cls):
		if cls._major_font is None:
			return ''
		return cls._major_font[0]

	@classmethod
	def list_font_source(cls):
		if cls._major_font is None:
			return ''
		return [(f[0], f[1]) for f in [cls._major_font] + cls._minor_font_list]

	@classmethod
	def decorate_text(cls, text):
		f = u'<span style="font-family:{}">{}</span>'.format
		def decorate_ch(decorated_text, ch):
			o = ord(ch)
			if o in cls._major_font[2]:
				return decorated_text + ch
			for name, file, width_map, default_width in cls._minor_font_list:
				if o in width_map:
					return decorated_text + f(name, ch)
			else:
				return decorated_text + ch

		return reduce(decorate_ch, text, '')

	@classmethod
	def get_text_width(cls, text, size):
		fonts = [cls._major_font] + cls._minor_font_list
		s = 0.001 * size
		def get_ch_width(lst, o):
			for name, file, width_map, default_width in fonts:
				if o in width_map:
					lst.append(s * width_map.get(o, default_width) + lst[-1])
					return lst
			else:
				lst.append(s * 500.0 + lst[-1])
				return lst

		return reduce(get_ch_width, [ord(ch) for ch in text], [0])

	@classmethod
	def get(cls, string_path, *args, **kwargs):
		if not string_path:
			return ''
		name_parts = string_path.split(':')
		ret = cls._table.get(cls.lang, cls._table['enu'])[name_parts[0]][name_parts[1]]
		if kwargs:
			for key, value in kwargs.items():
				ret = re.sub(r'\{%s\}' % (key,), lambda m: text_type(value), ret)
				ret = re.sub(r'\{%s:htmlEncode\}' % (key,), lambda m: escape(text_type(value)), ret)
		if args:
			return ret.format(*args)

		return ret

	@classmethod
	def get_log_string(cls, log_type, other_params, is_plain_text):
		return cls._log_ctx.SYNO.ActiveBackup.LogUtil.getLogString(log_type, other_params, is_plain_text)

	@classmethod
	def get_log_string_with_template(cls, log_type, other_params, is_plain_text, template_string):
		return cls._log_ctx.SYNO.ActiveBackup.LogUtil.getLogString(log_type, other_params, is_plain_text, template_string)

	@classmethod
	def get_log_template_string(cls, log_type):
		return cls._log_ctx.SYNO.SDS.ActiveBackup.getLogTemplateString(log_type)

	@classmethod
	def get_log_type_string(cls, log_level):
		if cls._log_ctx.SYNO.ActiveBackup.LogLevel.ERROR == log_level:
			return cls.get('log:error_type')
		elif cls._log_ctx.SYNO.ActiveBackup.LogLevel.WARNING == log_level:
			return cls.get('log:warning_type')
		elif cls._log_ctx.SYNO.ActiveBackup.LogLevel.INFO == log_level:
			return cls.get('log:info_type')
		else:
			return ''

	@classmethod
	def get_result_status_string(cls, status):
		return cls._log_ctx.SYNO.SDS.ActiveBackup.Utils.getResultStatusString(status)

	@classmethod
	def get_job_action_string(cls, action):
		return cls._log_ctx.SYNO.SDS.ActiveBackup.Utils.renderJobAction(action)

size_table = ['Byte', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
def get_unit(size):
	power = 0
	if size:
		power = int(math.log(size, 2) / 10.0)

	if power >= len(size_table):
		return size_table[-1], 2 ** (10 * (len(size_table) - 1))

	return size_table[power], float(2 ** (10 * power))

def add_months(sourcedate, months):
	month = sourcedate.month - 1 + months
	year = sourcedate.year + month // 12
	month = month % 12 + 1
	day = min(sourcedate.day, calendar.monthrange(year,month)[1])
	return datetime(year, month, day)

def generate_time_series(time_start, time_end, time_unit):
	pin = time_start
	ret = []
	idx = 0
	if time_unit == 'daily':
		while pin <= time_end:
			ret.append(pin)
			idx += 1
			pin = time_start + timedelta(hours=idx)
	elif time_unit in ('weekly', 'monthly'):
		while pin <= time_end:
			ret.append(pin)
			idx += 1
			pin = time_start + timedelta(days=idx)
	else:
		while pin <= time_end:
			ret.append(pin)
			idx += 1
			pin = add_months(time_start, idx)
	return ret

def to_timestamp(dt):
	return (dt - datetime.fromtimestamp(0)).total_seconds()
