#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2021 Synology Inc. All rights reserved.
import enum, errno, json, logging, logging.handlers, os, signal, six, socket, stat, subprocess, threading
import js2py
from util import load_js_as_json

if six.PY3:
	import queue as Queue
elif six.PY2:
	import Queue

SZ_SOCKET_PATH_SYSLOG_SERVICE = '/run/synoabk/synoabk-syslog-service.sock'
ABK_REPORT_TEMPLATE_DIR = '/var/packages/ActiveBackup/target/report_template/'
REQUEST_END_MAGIC = '^^2552*1814^^' # defined in ActiveBackup-core/lib/string_ipc.cpp
BUF_SIZE = 1024

if __name__ == '__main__':
	logger = logging.getLogger('activity_log_service')
	logger.setLevel(logging.DEBUG)

	handler = logging.handlers.SysLogHandler(address='/dev/log')
	handler.setLevel(logging.INFO)

	formatter = logging.Formatter('[%(levelname)s] synoabk_activity_log::%(message)s')
	handler.setFormatter(formatter)

	logger.addHandler(handler)

	class LogLevel(enum.Enum): # defined in ActiveBackup-core/include/synoabk/log_type.h
		ERROR     = 0
		WARNING   = 1
		INFO      = 2
		DEBUG     = 3

	log_queue = Queue.Queue()

	def StartWriteToLogLoop():
		lang_table = load_js_as_json(ABK_REPORT_TEMPLATE_DIR + 'lang.js')
		log_ctx = js2py.EvalJs()
		log_ctx.execute('''
			var window = this;
			function _LABS(sec, key) { return lang_table[sec][key] || ''; };
			var lang_table =''' + json.dumps(lang_table['enu']) + ';')
		with open(ABK_REPORT_TEMPLATE_DIR + 'pdf.js', 'r') as f:
			log_ctx.execute(f.read())
		del lang_table

		while True:
			try:
				json_data = json.loads(six.ensure_str(log_queue.get(), 'utf-8', 'ignore').replace(REQUEST_END_MAGIC, ''))
				if (json_data.get('action', '') != 'write_activity_log') or ('log' not in json_data):
					continue

				if json_data.get('log').get('other_params', {}):
					json_data.get('log').update(json_data.get('log').get('other_params', {}))

				log_string = log_ctx.SYNO.ActiveBackup.LogUtil.getLogString(int(json_data.get('log').get('log_type')), json_data.get('log'), True)
				if json_data.get('log').get('task_name', ''):
					log_string = '[' + json_data.get('log').get('task_name') + '] ' + log_string

				log_level = int(json_data.get('log').get('log_level', LogLevel.DEBUG.value))
				if log_level == LogLevel.ERROR.value:
					logger.error(log_string)
				elif log_level == LogLevel.WARNING.value:
					logger.warning(log_string)
				elif log_level == LogLevel.INFO.value:
					logger.info(log_string)
				else:
					logger.debug(log_string)
			except:
				continue

	def beUser(uid, gid):
		if os.getuid() != 0:
			return
		# Remove group privileges
		os.setgroups([])
		# Try setting the new uid/gid
		os.setgid(gid)
		os.setuid(uid)
		# Ensure a very conservative umask
		os.umask(0o77)

	def HandleSignal(sig, frame):
		if sig == signal.SIGPIPE:
			return
		os._exit(0)

	pkg_uid = int(subprocess.check_output("/usr/syno/sbin/synouser --get ActiveBackup | grep -i 'user uid' | grep -oE '[0-9]+'", shell=True))
	pkg_gid = int(subprocess.check_output("/usr/syno/sbin/synogroup --get ActiveBackup | grep -i 'group id' | grep -oE '[0-9]+'", shell=True))

	beUser(pkg_uid, pkg_gid)
	try:
		os.unlink(SZ_SOCKET_PATH_SYSLOG_SERVICE)
	except Exception:
		pass

	try:
		os.makedirs(os.path.dirname(SZ_SOCKET_PATH_SYSLOG_SERVICE))
	except OSError as e:
		if e.errno != errno.EEXIST:
			raise

	sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	sock.bind(SZ_SOCKET_PATH_SYSLOG_SERVICE)
	sock.listen(16)

	os.chown(SZ_SOCKET_PATH_SYSLOG_SERVICE, pkg_uid, pkg_gid)
	os.chmod(SZ_SOCKET_PATH_SYSLOG_SERVICE, stat.S_IRUSR + stat.S_IWUSR)

	for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGPIPE):
		signal.signal(sig, HandleSignal)

	threading.Thread(target=StartWriteToLogLoop).start()
	while True:
		try:
			connection, addr = sock.accept()
			data = connection.recv(BUF_SIZE)
			connection.sendall(b'{"success": true}')
		except:
			continue
		finally:
			connection.close()

		log_queue.put(data)
