#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

from future import standard_library
standard_library.install_aliases()
standard_library.install_hooks()

import xml.etree.ElementTree as ET
import os, traceback, socket, ssl, json, hashlib, winrm, stat
import ntpath, threading, time, re
import six
import errno
import paramiko
import requests
import codecs
from .defines import *
from .utils import fixListFormat
from paramiko.ssh_exception import SSHException
from http.client import BadStatusLine
from base64 import b64encode
from scp import SCPClient, SCPException
from func_timeout import func_timeout, FunctionTimedOut
from .codepage import CodePageTable

from .exceptions import (
    AuthenticateError,
    CerticateVerifiedError,
    EntityNotFound,
    VirtualDiskNotFound,
    GuestOSAttributeError,
    GuestOSException,
    InvalidArgument,
    OperationTimeoutError,
    PowerShellException,
    PathRestoreNotSupport,
    CodePageNotFound,
    Cancel
)

from .utils import (
    escapePath,
    toList,
    toUnixPath,
    toWinPath,
    predictRestoreMeta,
    getJsonFromString,
    Progress,
    exponentialRetryCall,
    PrivilegeEnv
)

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

import logging
import logging.handlers
logger = logging.getLogger(VM_LOGGER_NAME)

RUN_JOB_SCRIPT \
    = '$_job = [WMI]$job.Job\n' \
    + 'if ($job.ReturnValue -eq 4096) {\n' \
    +     'while ($_job.JobState -eq 4) {\n' \
    +         'Start-Sleep -seconds 1\n' \
    +         '$_job.PSBase.Get()\n' \
    +     '}\n' \
    +     'if ($_job.JobState -ne 7) {\n' \
    +         'if ($_job.ErrorDescription -ne "") {\n' \
    +             'Throw $_job.ErrorDescription\n' \
    +         '} else {\n' \
    +             'Throw "Error Code: $_job.ErrorCode, Job State: $_job.JobState"\n' \
    +         '}\n' \
    +     '}\n' \
    + '} elseif ($job.ReturnValue -ne 0) {\n' \
    +     'Throw "Error Code: $job.ReturnValue"\n' \
    + '}\n'

def runGuestShellBySSH(ssh, command, timeout=180, chcp=False):
    try:
        detect_codepage = "65001"
        if chcp:
            cpin, cpout, cperr = ssh.exec_command('chcp')
            detect_codepage = six.ensure_str(cpout.read(), 'utf-8-sig', 'ignore').split(':')[-1].strip()
            ssh.exec_command("chcp 65001")
        stdin, stdout, stderr = ssh.exec_command(command)
        if timeout > 0:
            polling_time = 0.1
            elapsed = 0
            while not stdout.channel.exit_status_ready():
                if elapsed >= timeout:
                    raise OperationTimeoutError('execute [%s] timeout' % command)
                time.sleep(polling_time)
                elapsed += polling_time
                if elapsed >= 1:
                    polling_time = 1

        status = stdout.channel.recv_exit_status()
        stdout = stdout.read().strip(b'\n')
        stderr = stderr.read().strip(b'\n')

        if chcp:
            try:
                stdout.decode('utf-8')
                stderr.decode('utf-8')
            except:
                if detect_codepage.isdigit() and int(detect_codepage) > 0:
                    codec = CodePageTable.getCodecByCodepage(int(detect_codepage))
                    stdout = stdout.decode(codec).encode('utf-8')
                    stderr = stderr.decode(codec).encode('utf-8')

        if status != 0:
            strerror = 'Failed to execute [%s] %s.' % (command, stderr)
            if status == 2147749891:
                strerror += ' Please make sure the user has administrative privilege.'
            raise GuestOSException(errno.EREMOTEIO, strerror)
    except (IOError, socket.gaierror) as e:
        raise GuestOSException(e.errno, e.strerror)
    except SSHException as e:
        strerror = six.u(str(e))
        if 'Channel closed.' == strerror:
            strerror = 'Executing commands on a remote machine over SSH is not available'
        raise GuestOSException(errno.ECONNRESET, strerror)

    return six.ensure_str(stdout)

class VmIO:
    def __init__(self, isWindowsVM, auth_info, overwrite = False, progressFilePath = None, parse_exception = None, logResultId = 0):
        if progressFilePath:
            self.progress = Progress(progressFilePath, parse_exception, logResultId)
        else:
            self.progress = None
        self.auth_info = auth_info
        self.is_windows_vm = isWindowsVM(auth_info['vm_id'], auth_info['username'], auth_info['password'], auth_info['address'], auth_info['port'])
        self.overwrite = overwrite
        self.sftp = None
        self.ssh = None
        self.transport = None
        self.temp_dst_path = None
        self.probily_meta = []
        self.cmdlineLimit = None
    def __del__(self):
        if self.transport:
            try:
                self.transport.close()
            except:
                logger.error(traceback.format_exc())
#-----------------------Private Function------------------------
    def _updateProgress(self, filename, total, transferred):
        if PrivilegeEnv().isCanceled():
            try:
                self._getSFTP().remove(self.temp_dst_path)
            except Exception:
                logger.error('remove temp. file [{}] failed.'.format(self.temp_dst_path))
            raise Cancel()

        self.progress.setProgress(transferred, 100.0 * transferred / total)

    def _getCmdlineLimit(self):
        if not self.is_windows_vm:
            return -1;
            #exec cmd immediately
        elif self.cmdlineLimit is not None:
            return self.cmdlineLimit

        ssh = self._getSSH()
        self.cmdlineLimit = 1900 #2047
        try:
            results = runGuestShellBySSH(ssh, 'ver', 180, True)
            if "Microsoft Windows" in results:
                #Microsoft Windows [version 10.0.18362.959]
                res = results.split(' ')
                res = res[-1][:-1] #10.0.18362.959
                versions = res.split('.')
                if versions[0].isdigit():
                    if int(versions[0]) >= 5:
                        self.cmdlineLimit = 8000 #8191
                        #https://support.microsoft.com/en-us/help/830473/command-prompt-cmd-exe-command-line-string-limitation
                        #Microsoft Windows XP or later,max length is 8191
                        #Other is 2047

                        #https://en.wikipedia.org/wiki/List_of_Microsoft_Windows_versions
                        #Microsoft Windows XP's version is 5.XXXX
            else:
                self.cmdlineLimit = -1
                #exec cmd immediately
        except:
            logger.error(traceback.format_exc())
            pass

        return self.cmdlineLimit

    def _windowBatchStatNode(self, checkList):
        if PrivilegeEnv().isCanceled():
            raise Cancel()

        results = []
        cmd = ''
        for dst_path in checkList:
            if cmd:
                cmd += " & "
            cmd += '(IF EXIST "%s" (IF EXIST "%s\\" (echo folder) ELSE (ECHO file)) ELSE (ECHO not exist))' % (toWinPath(dst_path), toWinPath(dst_path))

        res = runGuestShellBySSH(self._getSSH(), cmd, 180, True);
        records = res.split('\r\n')

        if len(checkList) != len(records):
            raise Exception("batch result not equal")

        for record in records:
            res = record.replace('\r','').replace('\n','')
            if res == 'folder':
                results.append(pathStatus.DIR)
            elif res == 'file':
                results.append(pathStatus.FILE)
            else:
                results.append(pathStatus.NOT_EXIST)
        return results


    def _batchCheckFileNeedUpload(self, checkList):
        if len(checkList) == 0:
            return []

        if PrivilegeEnv().isCanceled():
            raise Cancel()

        existList = []
        if self.is_windows_vm:
            records = []
            try:
                statList = []
                for (src_path, dst_path, temp_dst_path) in checkList:
                    statList.append(dst_path)
                records = self._windowBatchStatNode(statList)
            except Cancel as e:
                for (src_path, dst_path, temp_dst_path) in checkList:
                    self.progress.addErr(toWinPath(dst_path), e)
                raise e
            except Exception as e:
                logger.error(traceback.format_exc())
                for (src_path, dst_path, temp_dst_path) in checkList:
                    self.progress.addErr(toWinPath(dst_path), e)
                return []

            for idx in range(len(records)):
                res = records[idx]
                (src_path, dst_path, temp_dst_path) = checkList[idx]
                try:
                    if res == pathStatus.DIR:
                        raise GuestOSException(
                                errno.EISDIR,
                                'failed to overwrite file/folder. please make sure the file/folder property is the same.',
                                address=self.auth_info['address'])
                    elif res == pathStatus.FILE and not self.overwrite:
                        self.progress.addFileSkipSuccess(toWinPath(dst_path), os.lstat(src_path).st_size)
                    else:
                        existList.append((src_path, dst_path, temp_dst_path))
                except Exception as e:
                    logger.error(traceback.format_exc())
                    self.progress.addErr(toWinPath(dst_path), e)
        else:
            for (src_path, dst_path, temp_dst_path) in checkList:
                try:
                    try:
                        dst_stat = self._getSFTP().stat(dst_path).st_mode
                    except IOError as e:
                        if e.errno != errno.ENOENT:
                            raise
                        existList.append((src_path, dst_path, temp_dst_path))
                    else:
                        if not stat.S_ISREG(dst_stat):
                            raise GuestOSException(
                                errno.EISDIR,
                                'Failed to overwrite file/folder. Please make sure the file/folder property is the same.',
                                address=self.auth_info['address'])
                        elif not self.overwrite:
                            self.progress.addFileSkipSuccess(dst_path, os.lstat(src_path).st_size)
                        else:
                            existList.append((src_path, dst_path, temp_dst_path))
                except Cancel as e:
                    self.progress.addErr(dst_path, e)
                    raise e
                except Exception as e:
                    self.progress.addErr(dst_path, e)
        return existList

    def _batchUploadSameFolderData(self, checkList):
        if len(checkList) == 0:
            return []

        def need_retry_exception(e):
            if type(e) == socket.error:
                if e.args[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE, errno.ECONNRESET):
                    return True
            elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.exceptions.SSLError, requests.ConnectionError, requests.Timeout):
                return True
            elif type(e) == EOFError:
                return True
            return False

        @exponentialRetryCall(5, need_retry_exception)
        def _uploadData(uploadList, src_path, dst_path, temp_dst_path):
            try:
                self.temp_dst_path = temp_dst_path
                scp = SCPClient(self._getTransport(), progress=self._updateProgress)
                try:
                    scp.put(src_path, temp_dst_path)
                except IOError as e:
                    if e.errno == errno.ENOTSUP:
                        raise GuestOSException(
                            errno.ENOTSUP,
                            'Cannot read files on Data Deduplication-Enabled volume.',
                            address=self.auth_info['address'])
                    else:
                        raise GuestOSException(
                            errno.EIO,
                            'Cannot access the specified device, path, or file.',
                            address=self.auth_info['address'])
                except SCPException as e:
                    logger.error("put Operation error: %s", str(e))
                    if self.progress.info.current_percent == 100.0:
                        raise GuestOSException(
                            errno.ENOSPC,
                            'There is not enough space to upload the file. [%s]' % dst_path,
                            address=self.auth_info['address'])
                    raise
                finally:
                    scp.close()
                self._getSFTP().chmod(temp_dst_path, stat.S_IRWXU)
                uploadList.append((src_path, dst_path, temp_dst_path))
            except Exception as e:
                logger.error(traceback.format_exc())
                try:
                    if self.is_windows_vm:
                        runGuestShellBySSH(self._getSSH(), 'del "%s"' % toWinPath(temp_dst_path), 180 , True)
                    else:
                        self._getSFTP().remove(temp_dst_path)
                except Exception:
                    logger.error('remove temp. file [{}] failed.'.format(temp_dst_path))
                raise e

        uploadList = []
        for (src_path, dst_path, temp_dst_path) in checkList:
            try:
                self.progress.startFile(dst_path)
                _uploadData(uploadList, src_path, dst_path, temp_dst_path)
            except Cancel as e:
                self.progress.addErr((toWinPath(dst_path) if self.is_windows_vm else dst_path), e)
                raise e
            except Exception as e:
                self.progress.addErr((toWinPath(dst_path) if self.is_windows_vm else dst_path), e)
        return uploadList

    def _batchRename(self, checkList):
        if len(checkList) == 0:
            return []

        if PrivilegeEnv().isCanceled():
            raise Cancel()

        if self.is_windows_vm:
            try:
                try:
                    cmd = ''
                    last_dst_path = ''
                    for (src_path, dst_path, temp_dst_path) in checkList:
                        if cmd:
                            cmd += " & "
                        cmd += '(move /y "%s" "%s")' % (toWinPath(temp_dst_path), toWinPath(dst_path))
                        last_dst_path = dst_path
                    try:
                        res = runGuestShellBySSH(self._getSSH(), cmd, 180, True);
                    except exception as e:
                        logger.error(traceback.format_exc())
                        if 'the process cannot access the file because it is being used by another process.' in six.u(str(e)):
                            raise GuestOSException(
                                errno.ebusy,
                                'the destination file [%s] is being used by another process.' % ', '.join(dst_path for (src_path, dst_path, temp_dst_path) in checkList),
                                address=self.auth_info['address'])
                        raise e
                except Exception as e:
                    try:
                        cmd = ''
                        for (src_path, dst_path, temp_dst_path) in checkList:
                            self.progress.addErr(toWinPath(dst_path), e)
                            if cmd:
                                cmd += " & "
                            cmd += '(del "%s")' % (toWinPath(temp_dst_path))

                        res = runGuestShellBySSH(self._getSSH(), cmd, 180, True);
                    except Exception:
                        logger.error('remove temp. file [{}] failed.'.format(', '.join(dst_path for (src_path, dst_path, temp_dst_path) in checkList)))
                    raise e
            except Cancel as e:
                raise e
            except Exception:
                return []

            for (src_path, dst_path, temp_dst_path) in checkList:
                self.progress.addFileSuccess(toWinPath(dst_path), True)
            return checkList
        else:
            renameList = []
            for (src_path, dst_path, temp_dst_path) in checkList:
                try:
                    try:
                        if self.overwrite:
                            try:
                                self._getSFTP().remove(dst_path)
                            except IOError as e:
                                if e.errno != errno.ENOENT:
                                    raise
                        self._getSFTP().rename(temp_dst_path, dst_path)
                        self.progress.addFileSuccess(dst_path, True)
                        renameList.append((src_path, dst_path, temp_dst_path))
                    except Exception as e:
                        logger.error(traceback.format_exc())
                        try:
                            self._getSFTP().remove(temp_dst_path)
                        except exception:
                            logger.error('remove temp. file [{}] failed.'.format(temp_dst_path))
                        raise e
                except Cancel as e:
                    logger.error(traceback.format_exc())
                    self.progress.addErr(dst_path, e)
                    raise e
                except Exception as e:
                    logger.error(traceback.format_exc())
                    self.progress.addErr(dst_path, e)
            return renameList

    def _batchUploadSameFolderFile(self, batchPairList):
        if len(batchPairList) == 0:
            return
        checkList = []
        temp_dst_folder = ''

        #prepare src_path, dst_path, temp_path
        if self.is_windows_vm:
            (src_path, dst_path) = batchPairList[0]
            device_id = ntpath.splitdrive(dst_path)[0]
            temp_dst_folder = os.path.join(device_id, 'Synology')

        randomList = []
        for (src_path, dst_path) in batchPairList:
            s_path = six.ensure_str(src_path)
            d_path = toUnixPath(six.ensure_str(dst_path))
            temp_dst_path = ''

            #avoid same temp name
            randstr = randomString()
            while randstr in randomList:
                randstr = randomString()

            randomList.append(randstr)

            if self.is_windows_vm:
                temp_dst_path = os.path.join(temp_dst_folder, randstr + '.syno_abb_tmp')
            else:
                temp_dst_path = d_path + '_abb.' + randstr

            checkList.append((s_path, d_path, temp_dst_path))

        #stat, create temp folder, upload, rename, updateMeta
        checkList = self._batchCheckFileNeedUpload(checkList)
        if self.is_windows_vm and len(checkList) > 0:
            try:
                self._getSFTP().mkdir(temp_dst_folder)
            except: # can not distinguish "Already Exist" error
                pass

        checkList = self._batchUploadSameFolderData(checkList)
        checkList = self._batchRename(checkList)
        for (src_path, dst_path, tmp_path) in checkList:
            try:
                self._updataMeta(src_path, dst_path)
            except Exception as e:
                logger.error(traceback.format_exc())
                self.progress.addErr((toWinPath(dst_path) if self.is_windows_vm else dst_path), e)

        return checkList

    def _getTransport(self):
        if self.transport:
            if self.transport.is_active():
                return self.transport
            try:
                self.transport.close()
            except:
                logger.error(traceback.format_exc())

        username = self.auth_info['username']
        password = self.auth_info['password']
        address = self.auth_info['address']
        port = self.auth_info['port']

        try:
            try:
                transport = paramiko.Transport((address, port))
                transport.default_window_size = paramiko.common.MAX_WINDOW_SIZE
                transport.packetizer.REKEY_BYTES = pow(2, 40)
                transport.packetizer.REKEY_PACKETS = pow(2, 40)

                transport.connect(
                    username = username,
                    password = password)
                self.transport = transport
            except (IOError, socket.gaierror) as e:
                raise GuestOSException(e.errno, e.strerror, address=address)
            except SSHException as e:
                strerror = six.u(str(e))
                if 'Channel closed.' == strerror:
                    strerror = 'Executing commands on a remote machine over SSH is not available'
                raise GuestOSException(errno.ECONNRESET, strerror, address=address)
        except Exception as e:
            try:
                transport.close()
            except:
                logger.error(traceback.format_exc())
            raise e

        return self.transport

    def _getSSH(self):
        if self.ssh and self.ssh._transport.is_active():
            return self.ssh
        transport = self._getTransport()
        self.ssh = paramiko.SSHClient()
        self.ssh._transport = transport
        return self.ssh

    def _getSFTP(self):
        if self.sftp and self.sftp.sock.get_transport().is_active():
            return self.sftp
        transport = self._getTransport()
        self.sftp = paramiko.SFTPClient.from_transport(transport)
        # put file hang workaround
        self.sftp.ultra_debug = True
        return self.sftp

    def _listFiles(self, file_path):
        out = {'files': [], 'remaining': 0}

        if not file_path:
            if self.is_windows_vm:
                results = runGuestShellBySSH(self._getSSH(), 'wmic logicaldisk where drivetype=3 get DeviceID')

                drives = results.split()[1:]
                for drive in toList(drives):
                    out['files'].append({
                        'name': drive.lower(),
                        'path': drive.lower(),
                        'is_dir': True,
                        'mtime': 0
                    })
            else:
                out['files'].append({
                    'name': '/',
                    'path': '/',
                    'is_dir': True,
                    'mtime': 0
                })
        else:
            if self.is_windows_vm:
                win_file_path = toWinPath(file_path)
                if not win_file_path.endswith('\\'):
                    win_file_path += '\\'

                def _parse_data(files, is_dir):
                    skipDir = [".", ".."]
                    for f in files:
                        filename = f.replace('\r','').replace('\n','')
                        if not filename:
                            continue
                        if is_dir and filename in skipDir:
                            continue
                        out['files'].append({
                            'name': filename,
                            'path': win_file_path + filename,
                            'is_dir': is_dir,
                            'mtime': 0
                        })

                #list dir
                results = runGuestShellBySSH(self._getSSH(), 'dir /a:d/b "%s"' % win_file_path, 180 , True)
                _parse_data(results.split('\r\n'), True)
                try:
                    #list not dir
                    results = runGuestShellBySSH(self._getSSH(), 'dir /a:-d/b "%s"' % win_file_path, 180 , True)
                    _parse_data(results.split('\r\n'), False)
                except Exception as e:
                    logger.error("list file failed [%s]" % str(e))
                    #no file in folder
                    pass

            else:
                results = self._getSFTP().listdir_attr(file_path)

                for result in results:
                    out['files'].append({
                        'name': result.filename,
                        'path': file_path + '/' + result.filename,
                        'is_dir': stat.S_ISDIR(result.st_mode),
                        'mtime': result.st_mtime
                    })
        return out

    def _uploadFile(self, src_path, dst_path):
        batchList = []
        batchList.append((src_path, dst_path))
        res = len(self._batchUploadSameFolderFile(batchList)) > 0
        return res

    def _updataMeta(self, src_path, dst_path):
        src_path = six.ensure_str(src_path)
        dst_path = toUnixPath(six.ensure_str(dst_path))
        miss_attr_list = []

        src_stat = None

        if len(self.probily_meta) == 0:
            self.probily_meta = predictRestoreMeta(self.is_windows_vm, self.auth_info['username'])

        try:
            src_stat = os.lstat(src_path)
        except OSError as e:
            raise GuestOSAttributeError(
                e.errno,
                'Failed to stat source file [%s]' % src_path,
                miss_attr_list=self.probily_meta)
        except Exception as e:
            raise GuestOSAttributeError(
                errno.EIO,
                'Failed to stat source file [%s]' % src_path,
                miss_attr_list=self.probily_meta)

        try:
            self._getSFTP().utime(dst_path, (src_stat.st_atime, src_stat.st_mtime))
        except Exception:
            miss_attr_list.append(ATTR_ATIME)
            miss_attr_list.append(ATTR_MTIME)

        if not self.is_windows_vm:
            try:
                self._getSFTP().chmod(dst_path, src_stat.st_mode)
            except Exception:
                miss_attr_list.append(ATTR_PERMS)
            if self.auth_info['username'] == 'root':
                try:
                    self._getSFTP().chown(dst_path, src_stat.st_uid, src_stat.st_gid)
                except Exception:
                    miss_attr_list.append(ATTR_UID)
                    miss_attr_list.append(ATTR_GID)

        if len(miss_attr_list) > 0:
            raise GuestOSAttributeError(
                errno.EREMOTEIO,
                'update [%s] attribute [%s] failed' % (dst_path, ', '.join(x for x in miss_attr_list)),
                miss_attr_list=miss_attr_list)

    def _uploadDir(self, src_path, dst_path):
        self.progress.startFile(dst_path)

        def _predictCmdLine(old_predict, new_path):
            predict = old_predict
            if old_predict:
                predict += ' & '
            new_dst_path = toWinPath(toUnixPath(six.ensure_str(new_path)))
            predict += '(IF EXIST "%s" (IF EXIST "%s\\" (echo folder) ELSE (ECHO file)) ELSE (ECHO not exist))' % (new_dst_path, new_dst_path)
            return predict

        src_path = six.ensure_str(src_path)
        dst_path = toUnixPath(six.ensure_str(dst_path))
        folderList = []

        #file part
        batchList = []
        batchTotalSize = 0
        batchCmdLine = ''

        batchTotalSizeLimit = 128 * 1024**2 #128MB
        batchCountLimit = 64
        batchCmdlineLimit = self._getCmdlineLimit()

        for f in os.listdir(src_path):
            src_fullpath = os.path.join(src_path, f)
            dst_fullpath = os.path.join(dst_path, f)
            if os.path.islink(src_fullpath):
                log_dst_fullpath = toWinPath(dst_fullpath) if self.is_windows_vm else dst_fullpath
                self.progress.addErr(log_dst_fullpath, PathRestoreNotSupport("Unsupported link file type"))
            elif os.path.isfile(src_fullpath):
                if batchTotalSize >= batchTotalSizeLimit or len(batchList) >= batchCountLimit or len(_predictCmdLine(batchCmdLine, dst_fullpath)) >= batchCmdlineLimit:
                    self._batchUploadSameFolderFile(batchList)
                    batchList = []
                    batchTotalSize = 0
                    batchCmdLine = ''

                batchList.append((src_fullpath, dst_fullpath))
                batchTotalSize += os.lstat(src_fullpath).st_size
                batchCmdLine = _predictCmdLine(batchCmdLine, dst_fullpath)
            elif os.path.isdir(src_fullpath):
                folderList.append((src_fullpath, dst_fullpath))
            else:
                log_dst_fullpath = toWinPath(dst_fullpath) if self.is_windows_vm else dst_fullpath
                self.progress.addErr(log_dst_fullpath, PathRestoreNotSupport("Unsupported file type(%s)" % os.stat(src_fullpath).st_mode))

        self._batchUploadSameFolderFile(batchList)
        batchList = []
        batchTotalSize = 0
        batchCmdLine = ''
        #folder part
        for (src_fullpath, dst_fullpath) in folderList:
            log_dst_fullpath = toWinPath(dst_fullpath) if self.is_windows_vm else dst_fullpath
            try:
                is_create = self._createDirectory(dst_fullpath)
                self._uploadDir(src_fullpath, dst_fullpath)
                self.progress.addDirSuccess(log_dst_fullpath, is_create or self.overwrite)
                if is_create or self.overwrite:
                    self._updataMeta(src_fullpath, dst_fullpath)
            except Cancel as e:
                self.progress.addErr(log_dst_fullpath, e)
                raise e
            except Exception as e:
                self.progress.addErr(log_dst_fullpath, e)

    def _createUnixDirectory(self, dir_path):
        dir_path = toUnixPath(six.ensure_str(dir_path))
        dir_stat = None
        try:
            dir_stat = self._getSFTP().stat(dir_path).st_mode
        except IOError as e:
            if e.errno == errno.ENOENT:
                self._getSFTP().mkdir(dir_path)
                return True
            else:
                raise e
        else:
            if not stat.S_ISDIR(dir_stat):
                raise GuestOSException(
                    errno.ENOTDIR,
                    'Failed to overwrite file/folder. Please make sure the file/folder property is the same.',
                    address=self.auth_info['address'])
            return False

    def _createWinDirectory(self, dir_path):
        dir_path = toWinPath(six.ensure_str(dir_path))
        dst_basename = ntpath.basename(dir_path)
        parent_path = ntpath.dirname(dir_path)
        list_result = self._listFiles(parent_path)

        for f in list_result['files']:
            if f['name'] == dst_basename:
                if not f['is_dir']:
                    raise GuestOSException(
                        errno.ENOTDIR,
                        'Failed to overwrite file/folder. Please make sure the file/folder property is the same.',
                        address=self.auth_info['address'])
                return False

        runGuestShellBySSH(self._getSSH(), 'mkdir "%s"' % dir_path, 180, True)
        return True

    def _createDirectory(self, dir_path):
        if self.is_windows_vm:
            return self._createWinDirectory(dir_path)
        else:
            return self._createUnixDirectory(dir_path)

#-----------------------Public Function------------------------
    def updataMeta(self, src_path, dst_path):
        self._updataMeta(src_path, dst_path)

    def uploadFile(self, src_path, dst_path):
        try:
            self._uploadFile(src_path, dst_path)
        except BaseException as e:
            logger.error(traceback.format_exc())
            self.progress.addErr((toWinPath(dst_path) if self.is_windows_vm else toUnixPath(dst_path)), e)
        self.progress.end()

    def uploadDir(self, src_path, dst_path):
        try:
            is_create = self._createDirectory(dst_path)
            self._uploadDir(src_path, dst_path)
            self.progress.addDirSuccess((toWinPath(dst_path) if self.is_windows_vm else toUnixPath(dst_path)), is_create or self.overwrite)
            if is_create or self.overwrite:
                self._updataMeta(src_path, dst_path)
        except BaseException as e:
            logger.error(traceback.format_exc())
            self.progress.addErr((toWinPath(dst_path) if self.is_windows_vm else toUnixPath(dst_path)), e)
        self.progress.end()

    def listFiles(self, file_path):
        return self._listFiles(file_path)

    def createDirectory(self, dir_path):
        return self._createDirectory(dir_path)


class Main:
    def __init__(self, hostname, username, password, thumbprint, port, protocol, taskPool, hostType, progressFilePath, logResultId):
        self.auth_info = {
            'hostname'  : hostname,
            'username'  : username,
            'password'  : password,
            'thumbprint': thumbprint,
            'protocol'  : protocol,
            'port'      : port
        }
        self.protocol = None
        self.shell_id = None
        self.error = None
        self.lock = threading.Lock()
        self.lockNum = 0
        self.lockNumLock = threading.Lock()

        self.task_pool = taskPool
        self.host_type = hostType
        self.progressFilePath = progressFilePath
        self.logResultId = logResultId

    def __del__(self):
        if self.protocol:
            self.protocol.close_shell(self.shell_id)
        logger.info('disconnect %s' % self.auth_info['hostname'])

    def dispatch(self, command, argument):
        command_func = getattr(self, command)
        argument = argument or {}
        response = {}
        timeout_func = 0

        always_trust_cert = False
        if 'always_trust_cert' in argument:
            always_trust_cert = argument['always_trust_cert']
            argument.pop('always_trust_cert')

        if 'force_connect' in argument:
            argument.pop('force_connect')

        if 'mnt_src_path' in argument and 'mnt_dst_path' in argument:
            self.mnt_src_path = argument['mnt_src_path']
            self.mnt_dst_path = argument['mnt_dst_path']
            argument.pop('mnt_src_path')
            argument.pop('mnt_dst_path')

        if 'func_call_timeout' in argument:
            timeout_func = argument['func_call_timeout']
            argument.pop('func_call_timeout')
        else:
            timeout_func = 300

        try:
            func_timeout(300, self.connect, args=(always_trust_cert,))

            if 0 == timeout_func:
                response['data'] = command_func(**argument)
            else:
                response['data'] = func_timeout(timeout_func, command_func, kwargs=argument)
            response['success'] = True
        except BaseException as e:
            logger.error(traceback.format_exc())
            response = self._parse_exception(e)
            response['success'] = False

        return response

    def _parse_exception(self, e):
        error_class = type(e).__name__
        error_message = str(e.strerror) if hasattr(e, 'strerror') else str(e)
        error_params = {}

        for attr in dir(e):
            #https://stackoverflow.com/questions/47775772/why-accessing-an-objects-attribute-characters-written-throws-exception-while
            if attr == 'characters_written':
                continue
            t = type(getattr(e, attr)).__name__
            if (t == 'str' or t == 'unicode' or t == 'int') and not attr.startswith("__"):
                error_params[attr] = getattr(e, attr)

        if type(e) == socket.error:
            if e.args[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE, errno.ECONNRESET):
                error_class = 'ConnectionFail'
        elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.exceptions.SSLError, requests.ConnectionError, requests.Timeout):
            error_class = 'ConnectionFail'
        elif type(e) == FunctionTimedOut:
            error_message = 'WinRM operation timed out after %s seconds.' % ("N/A" if e.timedOutAfter == None else str(e.timedOutAfter))
        elif type(e) == GuestOSAttributeError:
            error_params['miss_attr_list'] = e.miss_attr_list

        return {
            'error_class': error_class,
            'error_message': error_message,
            'error_params': error_params
        }

    def _connect(self, always_trust_cert):
        if not self.auth_info['password']:
            raise AuthenticateError('Empty password')

        cert_validation = 'validate'
        ca_path = '/etc/ssl/certs'
        if True == always_trust_cert:
            cert_validation = 'ignore'
            ca_path = ''

        on_disk_thumbprint = self.auth_info['thumbprint']

        self.protocol = winrm.Protocol(
            endpoint='%s://%s:%d/wsman' % (self.auth_info['protocol'], self.auth_info['hostname'], self.auth_info['port']),
            transport='ntlm',
            username=self.auth_info['username'],
            password=self.auth_info['password'],
            ca_trust_path=ca_path,
            server_cert_validation=cert_validation)

        try:
            self.shell_id = self.protocol.open_shell(codepage=65001) # utf-8
        except requests.exceptions.SSLError as e:
            if True == always_trust_cert:
                raise requests.ConnectTimeout(e) #5395: SSL handshake timeout

            cert_pem = ssl.get_server_certificate((self.auth_info['hostname'], self.auth_info['port']))
            currentThumbprint = self.getServerCert()['thumbprint']
            if currentThumbprint != on_disk_thumbprint:
                reason = 'verify_failed'
                if "hostname" in str(e) and "doesn't match" in str(e):
                    reason = 'hostname_not_match'
                logger.error('certificate verified error!! except: %s, check_hostname: %s, on_disk_thumbprint: %s, current thumbprint: %s, always_trust_cert: %d, cert_validation: %s' % (str(e), self.auth_info['hostname'], on_disk_thumbprint, currentThumbprint, always_trust_cert, cert_validation))
                raise CerticateVerifiedError(reason, certificate=cert_pem.encode())

            self.protocol = winrm.Protocol(
                endpoint='%s://%s:%d/wsman' % (self.auth_info['protocol'], self.auth_info['hostname'], self.auth_info['port']),
                transport='ntlm',
                username=self.auth_info['username'],
                password=self.auth_info['password'],
                server_cert_validation='ignore')
            self.shell_id = self.protocol.open_shell(codepage=65001) # utf-8

        try:
            self.protocol.run_command(
                    self.shell_id,
                    'winrm set winrm/config/service \'@{MaxConcurrentOperationsPerUser="4294967295"}\'') # workaround
        except:
            logger.error(traceback.format_exc())
        logger.info('connect %s' % self.auth_info['hostname'])

    def connect(self, always_trust_cert):
        try:
            self.lockNumLock.acquire()
            self.lockNum += 1
            self.lockNumLock.release()
            self.lock.acquire()

            if self.error:
                raise self.error
            if self.protocol == None:
                self._connect(always_trust_cert)
        except Exception as e:
            self.error = e
            raise
        finally:
            self.lockNumLock.acquire()
            self.lockNum -= 1
            if self.lockNum == 0:
                self.error = None
            self.lockNumLock.release()
            self.lock.release()

    def checkHyperVAvailable(self, ignore_win_10 = True):
        data = {}
        data['success'] = True
        try:
            self._run_cmd('get-windowsfeature', False)
        except BaseException as e:
            if ignore_win_10:
                data['is_win_10'] = True
                return data

        try:
            data['hyperv_enabled'] = (self._run_cmd('(get-windowsfeature -Name "hyper-v").InstallState', False) == 'Installed')
            data['hyperv_ps_enabled'] = (self._run_cmd('(get-windowsfeature -Name "hyper-v-powershell").InstallState', False) == 'Installed')
            data['vmms_enabled'] = (self._run_cmd('(get-service vmms).status', False) == 'Running')
        except BaseException as e:
            data['success'] = False

        return data

    def checkFailoverCluster(self):
        data = {}
        data['cluster_enabled'] = True
        try:
            data['cluster_ip'] = self._run_cmd('(get-clusterresource | Where-Object { $_.ResourceType -eq "IP Address" } | get-clusterparameter -name Address).value', False)
            data['cluster_name'] = self._run_cmd('$clu = (get-cluster); $clu.name + \'.\' + $clu.domain', False)
        except BaseException as e:
            data['cluster_enabled'] = False

        return data

    def getServerInfo(self, reconnect=False):
        if reconnect:
            shell_id = self.protocol.open_shell()
            self.protocol.close_shell(shell_id)

        data = {}
        data['api_type'] = 'HyperV'
        data['host_version'] = self._run_cmd('(Get-WmiObject Win32_OperatingSystem).Caption', False)
        data['host_build_ver'] = self._run_cmd('(Get-WmiObject Win32_OperatingSystem).BuildNumber', False)
        data['instance_uuid'] = self._run_cmd('Get-Wmiobject Win32_ComputerSystemProduct | Select-Object -ExpandProperty UUID', False)
        data['licensed'] = True
        data['shell_available'] = True
        data['guest_available'] = True
        return data

    def getServerCert(self):
        if self.auth_info['protocol'] != 'https':
            return {'thumbprint': ''}

        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock = context.wrap_socket(sock, server_hostname=self.auth_info['hostname'])
        sock.connect((self.auth_info['hostname'], self.auth_info['port']))
        cert = sock.getpeercert(True)
        sock.close()

        sha1 = hashlib.sha1()
        sha1.update(cert)
        digest = sha1.hexdigest().lower()

        s, i = '', 0
        for c in digest:
            s += c
            i += 1
            if i % 2 == 0:
                s += ':'

        return {'thumbprint': s[:-1]}

    def getHostDefaultVersion(self):
        return self._run_cmd('Get-VMHostSupportedVersion -Default | Select-Object -ExpandProperty Version')

    def getHostSupportVersion(self):
        return self._run_cmd('Get-VMHostSupportedVersion | Select-Object -ExpandProperty Version')

    def getVmInfo(self, vm_id, get_detail=True):
        return self._parse_vm_data(vm_id, get_detail)

    def getClusterVm(self, vm_id):
        return self._run_cmd('Get-ClusterResource | Where ResourceType -eq "Virtual Machine"'
            + '| Select @{Name="Host";Expression={$_.OwnerNode.Name.ToUpper()}},'
            +          '@{Name="VmID";Expression={(Get-ClusterParameter -InputObject $_ | Where { $_.Name -eq "VmID" }).Value}}'
            + '| Where VmID -eq ' + vm_id)

    def getDatastoreInfo(self, datastore_infos):
        return [self._parse_datastore_data(datastore['datastore_id']) for datastore in datastore_infos]

    def listClusterSharedVolumePath(self):
        return [p.lower() for p in toList(self._run_cmd('(Get-ClusterSharedVolume).SharedVolumeInfo.FriendlyVolumeName'))]

    def getEntityPath(self, entityId=None, viewType=None, attr=None):
        if viewType == 'vim.Folder' or \
            viewType == 'vim.Datastore':
            if entityId == '':
                return ''
            if not self._run_cmd('Test-Path -LiteralPath "%s" -PathType Container' % escapePath(entityId)):
                raise EntityNotFound('Cannot find folder [%s]' % entityId)

            path = entityId.split('\\')
            path = [x for x in path if x]

            for i in range(1, len(path)):
                path[i] = path[i-1] + '\\' + path[i]
            return '/' + '/'.join(path)

        if viewType == 'vim.Network':
            if entityId == '':
                return ''
            self._run_cmd('Get-VMSwitch | Where {$_.Id.Guid -eq "%s"} | Select Name' % entityId)
            return '/' + entityId

        if viewType == 'vim.VirtualMachine-Host' or \
            viewType == 'vim.VirtualMachine-Folder' or \
            viewType == 'vim.VirtualMachine':
            self._run_cmd('Get-VM -Id %s | Select Name' % entityId)
            return '/' + entityId

        if viewType == 'vim.HostSystem' or \
            viewType == 'vim.ResourcePool':
            return ''

        raise InvalidArgument('invalid view type [%s]' % viewType)

    def listClusterNodeAddress(self):
        nodes = self._run_cmd('Get-ClusterNetworkInterface | select Node,Ipv4Addresses')
        nodes = fixListFormat(nodes)

        ip_list = {}
        for node in nodes:
            if node['Ipv4Addresses']:
                ip_list[node['Node']] = node['Ipv4Addresses'][0] 

        return ip_list

    def listChildEntity(self, parentId='', viewType='', getDetail=False):
        def _collect_vms(clustered=None):
            return self._parse_vm_data(None, getDetail, clustered)

        def _collect_drives(drives):
            entities = []
            for drive in toList(drives):
                entities.append({
                    'id': drive['DeviceID'].lower(),
                    'name': drive['DeviceID'].lower(),
                    'type': viewType
                })
            return entities

        def _collect_folders(folders):
            entities = []
            for folder in toList(folders):
                entities.append({
                    'id': folder['FullName'].lower(),
                    'name': folder['BaseName'].lower(),
                    'type': viewType
                })
            return entities

        def _collect_networks(networks):
            entities = []
            for network in toList(networks):
                entities.append({
                    'id': network['Id'],
                    'name': network['Name'],
                    'type': 'vim.Network'
                })
            return entities

        def _collect_hosts():
            hostName = self._run_cmd('$env:computername', False)
            return [{
                'id': hostName,
                'name': hostName,
                'type': 'vim.HostSystem',
                'migratable': self.getMigrateCapability()
            }]

        if parentId == '':
            if viewType == 'vim.Folder' or \
               viewType == 'vim.Datastore':
                drives = self._run_cmd('Get-WMIObject -Query "Select DeviceID From win32_logicaldisk Where DriveType=3"')
                entities = _collect_drives(drives)
            elif viewType == 'vim.Network':
                networks = self._run_cmd('Get-VMSwitch | Select Id,Name')
                entities = _collect_networks(networks)
            elif viewType == 'vim.VirtualMachine' or \
                viewType == 'vim.VirtualMachine-Host' or \
                viewType == 'vim.VirtualMachine-Folder':
                entities = _collect_vms()
            elif viewType == 'vim.ResourcePool':
                entities = _collect_hosts()
            else:
                entities = []
        else:
            if viewType == 'vim.Folder' or \
               viewType == 'vim.Datastore':
                folders = self._run_cmd('Get-ChildItem -Force -Directory -LiteralPath "%s/" | Where {$_.Attributes -notmatch "System" -or $_.Attributes -notmatch "Hidden"} | Select FullName,BaseName' % escapePath(parentId))
                entities = _collect_folders(folders)
            elif viewType == 'vim.VirtualMachine-IsCluster':
                entities = _collect_vms(clustered=True)
            elif viewType == 'vim.VirtualMachine-NotCluster':
                entities = _collect_vms(clustered=False)
            else:
                entities = []

        entities.sort(key=lambda obj: (obj['name']))
        entities.sort(key=lambda obj: (obj['type']), reverse=True)
        return entities

    def registerVm(self,
            vm_name=None,
            folder_id=None,
            resourcepool_id=None,
            host_id=None,
            os_name=None,
            cpu_number=None,
            memory_size=None,
            is_system_mbr=None,
            vdisks=None,
            networks=None,
            file_path=None,
            gen_new_uuid=False,
            to_cluster=False,
            secure_boot=False):

        cmd = '$vdiskPaths = @("' + '","'.join([escapePath(vdisk['filename']) for vdisk in vdisks]) + '")\n' \
            + '$vdiskUuids = @("' + '","'.join([vdisk['lunUuids'] for vdisk in vdisks]) + '")\n' \
            + '$vdiskIde = @(' + ','.join([str(int(vdisk['use_ide'])) for vdisk in vdisks]) + ')\n'

        if 0 < len(networks):
            cmd += '$networks = @("' + '","'.join([escapePath(network['name']) if network['name'] else '' for network in networks]) + '")\n'
        else:
            cmd += '$networks = @()\n'

        cmd += '$movedItems = @()\n' \
            + '$createdItems = @()\n' \
            + ('$vmName = "%s"\n' % escapePath(vm_name)) \
            + ('$cfgPath = "%s"\n' % escapePath(folder_id)) \
            + ('$vmxPath = "%s"\n' % escapePath(file_path)) \
            + 'for ($i=0; $i -lt $vdiskPaths.length; ++$i) {\n' \
            +     'if ($vdiskUuids[$i] -ne "") {\n' \
            +         'continue\n' \
            +     '}\n' \
            +     '$vdiskItem = Get-Item -LiteralPath $vdiskPaths[$i]\n' \
            +     '$diskHasData = Join-Path $vdiskItem.Directory ("abb_disk_has_data_" + $i)\n' \
            +     'if (-Not (Test-Path -LiteralPath $diskHasData)) {\n' \
            +         'New-Item $diskHasData | Out-Null\n' \
            +         '$createdItems += ,$diskHasData\n' \
            +         'continue\n' \
            +     '}\n' \
            +     '$newFolder = Split-Path $vdiskItem.Directory -Parent\n' \
            +     '$newBaseName = $vdiskItem.Basename\n' \
            +     'for ($num = 1; Test-Path -LiteralPath (Join-Path $newFolder ($newBaseName + $vdiskItem.Extension)); ++$num) {\n' \
            +         '$newBaseName = $vdiskItem.Basename + "_$num"\n' \
            +     '}\n' \
            +     '$vdiskPaths[$i] = Join-Path $newFolder ($newBaseName + $vdiskItem.Extension)\n' \
            +     'Move-Item -LiteralPath $vdiskItem.FullName $vdiskPaths[$i]\n' \
            +     '$movedItems += ,@($vdiskItem.FullName, $vdiskPaths[$i])\n' \
            + '}\n'

        vmxExisted = self._run_cmd('Test-Path -LiteralPath "%s"' % escapePath(file_path))
        if vmxExisted:
            genMacCmd \
                = 'if ($config.VM.NetworkAdapters[$i].DynamicMacAddressEnabled -eq $false) {\n' \
                +     '$newMacAddress = "00" + ((0..9 | %{"{0:x}" -f (Get-Random -Minimum 0 -Maximum 15)}) -join "")\n' \
                +     'Set-VMNetworkAdapter -VMNetworkAdapter $config.VM.NetworkAdapters[$i] -StaticMacAddress $newMacAddress\n' \
                + '}\n'
            addDiskCmd = 'Add-VMHardDiskDrive -VM $vm -ControllerType $hardDrives[$j].ControllerType -ControllerNumber $hardDrives[$j].ControllerNumber -ControllerLocation $hardDrives[$j].ControllerLocation'
            addDiskInfoCmd \
                = 'if ($vdiskUuids[$i] -ne "") {\n' \
                +     '$disk = Get-WmiObject Win32_DiskDrive | Where SerialNumber -eq $vdiskUuids[$i] | Select -ExpandProperty Index\n' \
                +     '$cmd = @"\n' \
                +     'select disk $disk\n' \
                +     'offline disk\n' \
                +     '"@\n' \
                +     '$cmd | diskpart | Out-Null\n' \
                +      addDiskCmd + ' -DiskNumber $disk\n' \
                + '} else {\n' \
                +      addDiskCmd + ' -Path $vdiskPaths[$i]\n' \
                + '}\n'

            cmd += '$vhdTempDirPath = (get-item $vmxPath).Directory.FullName\n'
            cmd += '$config = Compare-VM -Copy -GenerateNewId -Path $vmxPath -VirtualMachinePath $cfgPath -SnapshotFilePath $cfgPath -SmartPagingFilePath $cfgPath -VhdDestinationPath $vhdTempDirPath \n' \
                + '$hardDrives = $config.VM.HardDrives\n' \
                + '$filterHardDrives = $hardDrives | Where Path\n' \
                + 'Rename-VM -VM $config.VM -NewName $vmName\n' \
                + 'try { Set-VM -VM $config.VM -AutomaticCheckpointsEnabled $false } catch {}\n' \
                + 'Set-VMMemory -VM $config.VM -StartupBytes ' + str(memory_size) + 'MB\n' \
                + 'Set-VMProcessor -VM $config.VM -Count ' + str(cpu_number) + '\n' \
                + 'Get-VMSnapshot $config.VM | %{ try { Remove-VMSnapshot $_ } catch {} }\n' \
                + 'Remove-VMHardDiskDrive -VMHardDiskDrive $config.VM.HardDrives\n' \
                + 'for ($i=0; $i -lt $networks.length; ++$i) {\n' \
                +     'if ($networks[$i]) {\n' \
                +         '$switch = Get-VMSwitch | Where Name -eq $networks[$i]\n' \
                +         'Connect-VMNetworkAdapter -VMNetworkAdapter $config.VM.NetworkAdapters[$i] -VMSwitch $switch\n' \
                +     '} else {\n' \
                +         'Disconnect-VMNetworkAdapter -VMNetworkAdapter $config.vm.NetworkAdapters[$i]\n' \
                +     '}\n' \
                +      (genMacCmd if gen_new_uuid else '') \
                + '}\n' \
                + '$test = Compare-VM -CompatibilityReport $config\n' \
                + '$incompatibilities = $test.incompatibilities | Where-Object {$_.MessageID -ne 40010} \n' \
                + 'if ($incompatibilities) { throw "New VM failed" } \n' \
                + '$vm = Import-VM -CompatibilityReport $config\n' \
                + '$pPath = split-path -Path $vmxPath\n' \
                + '$ctrlNotePath = Join-Path $pPath "ctrl.tmp"\n' \
                + 'if (Test-Path $ctrlNotePath -PathType Leaf) {\n' \
                +     '$diskDist = @()\n' \
                +     '$diskDist += get-content $ctrlNotePath | ConvertFrom-Json\n' \
                +     'for ($j=0; $j -lt $hardDrives.length; ++$j) {\n' \
                +         'for ($i=0; $i -lt $diskDist.length; ++$i) {\n' \
                +             'if (($diskDist[$i].type -eq $hardDrives[$j].ControllerType) -And ($diskDist[$i].ctrl.ControllerNumber -eq $hardDrives[$j].ControllerNumber) -And ($diskDist[$i].unit -eq $hardDrives[$j].ControllerLocation)) {\n' \
                +                 addDiskInfoCmd \
                +                 'break\n' \
                +             '}\n' \
                +         '}\n' \
                +     '}\n' \
                + '} else {\n' \
                +     '$hardDrives = $filterHardDrives\n' \
                +     'for ($i=0; $i -lt $hardDrives.length; ++$i) {\n' \
                +          '$j = $i\n' \
                +          addDiskInfoCmd \
                +     '}\n' \
                + '}\n' \
                + 'if ($vm.Generation -eq 1) {\n' \
                +     'Set-VMBios -VM $vm -StartupOrder @("IDE", "CD", "LegacyNetworkAdapter", "Floppy")\n' \
                + '} else {\n' \
                +     'Get-VMHardDiskDrive -VM $vm | %{\n' \
                +         'Set-VMFirmware -VM $vm -FirstBootDevice $_\n' \
                +     '}\n' \
                + '}\n'
        else:
            addDiskCmd = 'Add-VMHardDiskDrive -VMDriveController $ctrl -ControllerLocation $unit'

            cmd += '$generation = ' + ('1' if is_system_mbr else '2') + '\n' \
                + '$vm = New-VM -Name $vmName -Generation $generation -Path $cfgPath\n' \
                + 'Set-VMMemory -VM $vm -StartupBytes ' + str(memory_size) + 'MB\n' \
                + 'Set-VMProcessor -VM $vm -Count ' + str(cpu_number) + '\n' \
                + 'try { Set-VM -VM $vm -AutomaticCheckpointsEnabled $false } catch {}\n' \
                + ('' if 'windows' in os_name.lower() else 'try {'\
                +       'Set-VMFirmware -VM $vm -EnableSecureBoot ' + ('On' if secure_boot else 'Off') \
                +               ' -SecureBootTemplate MicrosoftUEFICertificateAuthority } catch {}\n') \
                + 'Remove-VMNetworkAdapter -VM $vm\n' \
                + 'Get-VMDvdDrive -VMName $vmName -ControllerNumber 0 | Remove-VMDvdDrive\n' \
                + 'Get-VMDvdDrive -VMName $vmName -ControllerNumber 1 | Remove-VMDvdDrive\n' \
                + '$ideCnt = 0\n' \
                + '$scsiCnt = 0\n' \
                + '$ctrlObject = @()\n' \
                + 'for ($i=0; $i -lt $vdiskPaths.length; ++$i) {\n' \
                +     'if ($vdiskIde[$i]) {\n' \
                +         '$ctrl = (Get-VMIdeController -VM $vm)[[Math]::floor($ideCnt / 2)]\n' \
                +         '$unit = $ideCnt % 2\n' \
                +         '$ideCnt += 1\n' \
                +         '$ctrlObject += @{"type" = "IDE"; "unit" = $unit; "ctrl" = $ctrl}\n' \
                +     '} else {\n' \
                +         '$ctrl = Get-VMScsiController -VM $vm\n' \
                +         '$unit = $scsiCnt\n' \
                +         '$scsiCnt += 1\n' \
                +         '$ctrlObject += @{"type" = "SCSI"; "unit" = $unit; "ctrl" = $ctrl}\n' \
                +     '}\n' \
                +     'if ($vdiskUuids[$i] -ne "") {\n' \
                +         '$disk = Get-WmiObject Win32_DiskDrive | Where SerialNumber -eq $vdiskUuids[$i] | Select -ExpandProperty Index\n' \
                +         '$cmd = @"\n' \
                +         'select disk $disk\n' \
                +         'offline disk\n' \
                +         '"@\n' \
                +         '$cmd | diskpart | Out-Null\n' \
                +          addDiskCmd + ' -DiskNumber $disk\n' \
                +     '} else {\n' \
                +          addDiskCmd + ' -Path $vdiskPaths[$i]\n' \
                +     '}\n' \
                + '}\n' \
                + 'for ($i=0; $i -lt $networks.length; ++$i) {\n' \
                +     '$switch = Get-VMSwitch | Where Name -eq $networks[$i]\n' \
                +     'Add-VMNetworkAdapter -VM $vm\n' \
                +     'Connect-VMNetworkAdapter -VMNetworkAdapter $vm.NetworkAdapters[$i] -VMSwitch $switch\n' \
                + '}\n' \
                + '$pPath = split-path -Path $vmxPath\n' \
                + '$ctrlNotePath = Join-Path $pPath "ctrl.tmp"\n' \
                + '$ctrlObject | ConvertTo-Json | Out-File $ctrlNotePath\n' \
                + '$specPath = $vmxPath -replace (".vmcx$", "")\n' \
                + '$specItem = Get-ChildItem -LiteralPath "$($vm.Path)\\Virtual Machines" -Filter ($vm.Id.Guid + ".*") -File\n' \
                + '$specItem | %{ Copy-Item -LiteralPath $_.FullName ($specPath + $_.Extension); $createdItems += ,($specPath + $_.Extension) }\n' \
                + '$test = Compare-VM -Copy -GenerateNewId -Path $vmxPath\n' \
                + '$incompatibilities = $test.incompatibilities | Where-Object {$_.MessageID -ne 40010} \n' \
                + 'if ($incompatibilities) { throw "New VM failed" } \n'

        if to_cluster:
            cmd += '$vm | Add-ClusterVirtualMachineRole | Out-Null\n'

        cmd = 'try {\n' \
            +     cmd + '$vm.Id.Guid\n' \
            + '} catch {\n' \
            +     'try { Remove-ClusterGroup -VMId $vm.Id -Force -RemoveResources } catch {}\n' \
            +     'try { Remove-VM -VM $vm -Force } catch {}\n' \
            +     'try { Remove-Item -LiteralPath $createdItems } catch {}\n' \
            +     'try { $movedItems | %{ Move-Item -LiteralPath $_[1] $_[0] }} catch {}\n' \
            +     '$incompatibilities | ConvertTo-Json \n' \
            +     'throw\n' \
            + '}'

        return self._run_cmd(cmd, False, True, True)

    def createDatastore(self, name=None, host_id=None, ip_address=None, mount_path=None):
        dst_root = self._run_cmd('echo $env:LOCALAPPDATA/ActiveBackup', False)
        dst_path = ntpath.join(dst_root, name + '-' + randomString())
        self.createFolder(folder_path=dst_path)
        return self._parse_datastore_data(dst_path)

    def removeDatastore(self, host_id=None, datastore_id=None):
        self.removeFolder(folder_path=datastore_id)

    def createFolder(self, datacenter_id=None, folder_path=None):
        self._run_cmd('New-Item -ItemType Directory -Path "%s"' % escapePath(folder_path), False)

    def removeFolder(self, datacenter_id=None, folder_path=None):
        self._run_cmd('Remove-Item -LiteralPath "%s" -Recurse -Force' % escapePath(folder_path), False)

    def renameFolder(self, datacenter_id=None, src_path=None, dst_path=None):
        self._run_cmd('Rename-Item -LiteralPath "%s" -NewName "%s"' % (escapePath(src_path), escapePath(dst_path)), False)

    def createDisk(self, file_path=None, thin_provision=None, capacity=None):
        disk_type = 'dynamic' if thin_provision else 'fixed'
        if capacity < 3145728:
            capacity = 3145728
        return self.runTask('New-VHD -%s -SizeBytes %s -Path "%s"' % (disk_type, capacity, escapePath(file_path)))

    def testConnection(self, ip=None, port=None):
        cmd = 'New-Object System.Net.Sockets.TCPClient -ArgumentList %s, %d | Select Connected'
        try:
            return self._run_cmd(cmd % (ip, port))["Connected"]
        except:
            return False

    def isIscsiRunning(self):
        return self._run_cmd('(Get-Service msiscsi).Status -eq "Running"')

    def isClusterEnabled(self):
        data = {}
        data['success'] = True
        try:
            data['cluster_ip'] = self._run_cmd('(get-clusterresource -name "cluster ip address" | get-clusterparameter -name Address).value', False)
            data['cluster_name'] = self._run_cmd('$clu = (get-cluster); $clu.name + \'.\' + $clu.domain', False)
        except BaseException as e:
            data['success'] = False

        return data

    def isSamba1Enabled(self):
        return self._run_cmd('Get-SmbServerConfiguration | Select -ExpandProperty EnableSMB1Protocol')

    def isSamba2Enabled(self):
        return self._run_cmd('Get-SmbServerConfiguration | Select -ExpandProperty EnableSMB2Protocol')

    def checkInvalidShares(self):
        all_drives = self._run_cmd('$a=""; [System.IO.DriveInfo]::GetDrives() | Where DriveType -Ne "CDROM" | %{ $a+=$_.Name[0] }; $a', False)
        net_drives = self._run_cmd('$a=""; Get-SmbShare | Where { $_.Name -Match \'^\\S\\$$\' } | %{ $a+=$_.Name[0] }; $a', False)
        return list(set(all_drives.upper()) - set(net_drives.upper()) - set('AB'))

    def isScriptCanRun(self):
        return not self._run_cmd('Try { $p="$env:temp/ActiveBackup-' + randomString() + '.ps1"; New-Item $p  -ItemType "file" -Value " " | Out-Null; &$p } Catch { $_ } Finally { Remove-Item (Get-Item $p).FullName }', False)

    def isPartOfDomain(self):
        try:
            self._run_cmd('Get-Cluster', False)
            return True
        except:
            return False

    def createIscsiPortal(self, portal_addr=None):
        try:
            self._run_cmd('Get-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' | Update-IscsiTargetPortal', True)
            return False
        except PowerShellException:
            self._run_cmd('New-IscsiTargetPortal -TargetPortalAddress ' + portal_addr, False)
        return True

    def updateIscsiTargetRemovePortal(self, portal_addr=None, remove_portal=False):
        cmd = 'Get-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' | Update-IscsiTargetPortal'
        if remove_portal:
            cmd += '\nRemove-IscsiTargetPortal -TargetPortalAddress ' + portal_addr + ' -Confirm:$false'

        self.runProcess(cmd, True)

    def createIscsiDisk(self, node_addr=None):
        try:
            self._run_cmd('Connect-IscsiTarget -NodeAddress ' + node_addr, False)
        except PowerShellException as e:
            if 0 > e.error_id.find('0xefff003f'):
                raise e

    def removeIscsiDisk(self, node_addr=None):
        cmd = '$ErrorActionPreference = "stop"\n' \
            + 'for ($i=0; $i -lt 10; ++$i) {\n' \
            +     'try {\n' \
            +         'Disconnect-IscsiTarget -NodeAddress ' + node_addr + ' -Confirm:$false\n' \
            +         'break\n' \
            +     '} catch {\n' \
            +         'sleep ([Math]::pow(2,$i))\n' \
            +     '}\n' \
            + '}'
        self.runProcess(cmd, True)

    def migrateVm(self, vm_id=None, datastore_id=None):
        pass

    def poweronVm(self, vm_id=None):
        self._run_cmd('Start-VM -VM (Get-VM -Id %s)' % vm_id)

    def poweroffVm(self, vm_id=None):
        self._run_cmd('Stop-VM -VM (Get-VM -Id %s) -TurnOff' % vm_id)

    def unregisterVm(self, vm_id=None):
        try:
            self._run_cmd('Remove-ClusterGroup -VMId %s -Force -RemoveResources' % vm_id)
        except:
            logger.debug(traceback.format_exc())

        self._run_cmd('Remove-VM -VM (Get-VM -Id %s) -Force' % vm_id)

    def listSnapshot(self, vm_id=None):
        return self._parse_snapshot_data(vm_id)

    def getSnapshot(self, vm_id=None, snap_id=None):
        return self._parse_snapshot_data(vm_id, snap_id)

    def takeSnapshot(self, vm_id=None, snap_name=None, snap_desc=None, quiesce=False):
        for snapshot in self.listSnapshot(vm_id):
            if snap_name == snapshot['name']:
                return snapshot

        vm = self._run_cmd('Get-VM -Id ' + vm_id + '| Select OperationalStatus,StatusDescriptions')
        if len(vm['OperationalStatus']) != 1 or vm['OperationalStatus'][0] != 2: # status != OK
            raise Exception('Virtual Machine is in busy status "%s".' % vm['StatusDescriptions'][-1])

        cmd = '$service = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSnapshotService\n' \
            + '$vm = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_ComputerSystem -Filter "Name=\'' + vm_id + '\'"\n' \
            + '$setting = ([WMIClass]"\\\\.\\root\\virtualization\\v2:Msvm_VirtualSystemSnapshotSettingData").CreateInstance()\n' \
            + '$setting.ConsistencyLevel = ' + ('1' if quiesce else '2') + '\n' \
            + '$setting.IgnoreNonSnapshottableDisks = $true\n' \
            + '$setting.GuestBackupType = 1\n' \
            + '$job = $service.CreateSnapshot($vm, $setting.GetText(2), 32768)\n' \
            + RUN_JOB_SCRIPT \
            + '$snapshot = ($_job.GetRelated("Msvm_VirtualSystemSettingData") | %{$_})\n' \
            + '$snapshotId = $snapshot.InstanceId.Split(":")[1]\n' \
            + 'Rename-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id $snapshotId) -NewName "' + snap_name + '"\n' \
            + '$snapshotId'

        try:
            snap_id = self._run_cmd(cmd, False, True, True)
            return self._parse_snapshot_data(vm_id, snap_id)
        except Exception as e:
            try:
                self.removeSnapshot(vm_id, snap_id)
            except:
                pass
            raise e

    def removeSnapshot(self, vm_id=None, snap_id=None, nonblocking=None):
        if nonblocking:
            self.runProcess('Remove-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id %s)' % snap_id, True)
            return None

        snapshot = self._parse_snapshot_data(vm_id, snap_id)

        if snapshot['type'] != 1: # recovery
            self._run_cmd('Remove-VMSnapshot -VMSnapshot (Get-VMSnapshot -Id %s)' % snap_id, False)
        else:
            cmd = '$key2Uuid = @{}\n' \
                + '$uuid2cbt = @{}\n' \
                + 'Get-VM -Id ' + vm_id + ' | Get-VMHardDiskDrive | %{\n' \
                +     'Try {\n' \
                +         '$path = ($_|Get-VHD).Path\n' \
                +         'while ($path) {\n' \
                +             '$disk = Get-VHD -Path $path\n' \
                +             '$path = $disk.ParentPath\n' \
                +         '}\n' \
                +         '$key2Uuid[($_.Id -Replace ".$")] = $disk.DiskIdentifier\n' \
                +     '} Catch {}\n' \
                + '}\n' \
                + '$snapshot = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSettingData -Filter "InstanceID=\'Microsoft:' + snap_id + '\'"\n' \
                + '$service = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_VirtualSystemSnapshotService\n' \
                + '$job = $service.ConvertToReferencePoint($snapshot)\n' \
                + RUN_JOB_SCRIPT \
                + '$refpoint = ($_job.GetRelated("Msvm_VirtualSystemReferencePoint") | %{$_})\n' \
                + 'for ($i = 0; $i -lt $refpoint.VirtualDiskIdentifiers.Count; $i++) {\n' \
                +     'Try {\n' \
                +         '$key = $refpoint.VirtualDiskIdentifiers[$i] -Replace ".$"\n' \
                +         '$cbt = $refpoint.ResilientChangeTrackingIdentifiers[$i]\n' \
                +         '$uuid2cbt[$key2Uuid[$key]] = $cbt\n' \
                +     '} Catch {}\n' \
                + '}\n' \
                + '$uuid2cbt'

            uuid2cbt = self._run_cmd(cmd, True, True, True)
            for vdisk in snapshot['virtual_disk']:
                try:
                    vdisk['change_id'] = uuid2cbt[vdisk['key']]
                except:
                    pass
        return snapshot

    def runTask(self, cmd=None):
        cmdJob = '$job = ' + cmd + ' -asJob\n' \
            + '$file = Join-Path $env:TEMP ("ActiveBackupTask_" + $PID)\n' \
            + 'While (($job.JobStateInfo.state -eq 0) -or ($job.JobStateInfo.state -eq 1)) {\n' \
            + '     $workaround = $job | Select *\n' \
            + '     $job.progress | ConvertTo-Json > $file\n' \
            + '     $job | Wait-Job -Timeout 1\n' \
            + '}\n' \
            + '$job.Error | ConvertTo-Json > $file'

        return self.runProcess(cmdJob, True)

    def removeAllSnapshots(self, vm_id=None, nonblocking=None):
        return None

    def runProcess(self, cmd=None, is_ps_script=False):
        if is_ps_script:
            sha1 = hashlib.sha1()
            sha1.update(six.ensure_binary(cmd))
            script = '$env:temp/synology-' + sha1.hexdigest() + '.ps1'
            cmd = 'echo  \'' + cmd + '\' > ' + script + '\n' \
                + 'Invoke-WmiMethod -Path win32_process -Name create -ArgumentList "powershell ' + script + '"'
        else:
            cmd = "Invoke-WmiMethod -Path win32_process -Name create -ArgumentList '%s'" % cmd
        return self._run_cmd(cmd)['ProcessId']

    def getTaskInfo(self, task_id=None):
        if task_id in self.task_pool:
            task = self._get_task(task_id)

            if task.info.state == 'error':
                self._del_task(task_id)
                raise task.info.error

            if task.info.state == 'success':
                self._del_task(task_id)

            return {
                'success': task.info.state == 'success',
                'progress': task.info.progress,
                'result': task.info.result
            }
        else:
            cmd = '$file = Join-Path $env:TEMP ActiveBackupTask_' + task_id + '\n'\
                + '$result = @{}\n' \
                + '$proc = Get-Process -Id ' + task_id + ' -ErrorAction SilentlyContinue\n' \
                + 'if ($proc) {\n' \
                + '     $result["success"] = $FALSE;\n' \
                + '     if (Test-Path $file) {\n' \
                + '         $progress = Get-Content $file | Out-String | ConvertFrom-Json\n' \
                + '         $result["progress"] = $progress.PercentComplete;\n' \
                + '     } else {\n' \
                + '         $result["progress"] = 0;\n' \
                + '     }\n' \
                + ' } else {\n' \
                + '     $result["success"] = $TRUE;\n' \
                + '     $result["progress"] = 100;\n' \
                + '     if (Test-Path $file) {\n' \
                + '         $progress = Get-Content $file | Out-String | ConvertFrom-Json\n' \
                + '         Remove-Item (Get-Item $file).FullName\n' \
                + '         if ($progress.Exception) {\n' \
                + '             $result["exception"] = $progress\n' \
                + '         }\n' \
                + '     }\n' \
                + ' }\n' \
                + '$result'

            ret = self._run_cmd(cmd, True, True)
            if 'exception' in ret:
                errmsg = ret['exception']
                excption = PowerShellException(six.ensure_str(json.dumps(errmsg), 'unicode-escape', 'ignore'))
                try:
                    excption = PowerShellException(six.ensure_str(json.dumps(errmsg['Exception']['Message']), 'unicode-escape', 'ignore'),
                        category_info=errmsg['CategoryInfo'],
                        error_id=errmsg['FullyQualifiedErrorId'])
                except:
                    pass

                raise excption
            return ret

    def cancelTask(self, task_id=None):
        if task_id in self.task_pool:
            task = self._get_task(task_id)
            task.CancelTask()
            self._del_task(task_id)
        else:
            self._run_cmd('Stop-Process -Id %s' % task_id)

    def updataMetaInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, is_folder=False, isWindows=False):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        auth_info = self._getAuthenInfo(vm_id, vm_user, vm_pwd, vm_address, vm_port)
        vmio = VmIO(self.isWindowsVM, auth_info)

        self.runGuest(vmio.updataMeta, auth_info, src_path, dst_path)

    def uploadDirInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, overwrite=True):
        if not len(self.progressFilePath):
            raise Exception('should provide file path for writing progress')

        auth_info = self._getAuthenInfo(vm_id, vm_user, vm_pwd, vm_address, vm_port)
        vmio = VmIO(self.isWindowsVM, auth_info, overwrite, self.progressFilePath, self._parse_exception, self.logResultId)
        return self.runGuest(vmio.uploadDir, auth_info, src_path, dst_path)

    def uploadFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None, overwrite=True):
        if not len(self.progressFilePath):
            raise Exception('should provide file path for writing progress')

        auth_info = self._getAuthenInfo(vm_id, vm_user, vm_pwd, vm_address, vm_port)
        vmio = VmIO(self.isWindowsVM, auth_info, overwrite, self.progressFilePath, self._parse_exception, self.logResultId)
        return self.runGuest(vmio.uploadFile, auth_info, src_path, dst_path)


    def isWindowsVM(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }
        isWindows = False
        try:
            results = self.runGuestShell(auth_info, 'ver', 180, True)
            if "Microsoft Windows" in results:
                isWindows = True
        except CodePageNotFound:
            raise
        except:
            #logger.error(traceback.format_exc())
            self.runGuestShell(auth_info, 'uname', 180)
            isWindows = False
        return isWindows

    def _getAuthenInfo(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }
        auth_info['address'], auth_info['port'], auth_info['username'], auth_info['password'] = self._get_vm_connection_info(auth_info)
        return auth_info

    def listFilesInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, index=None, max_results=None, match_pattern=None):
        auth_info = self._getAuthenInfo(vm_id, vm_user, vm_pwd, vm_address, vm_port)
        vmio = VmIO(self.isWindowsVM, auth_info)
        return self.runGuest(vmio.listFiles, auth_info, file_path)

    def downloadFileFromVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, src_path=None, dst_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        def func(transport, src_path, dst_path):
            scp = SCPClient(transport)
            src_path = six.ensure_str(src_path)
            dst_path = six.ensure_str(dst_path)
            scp.get(toUnixPath(src_path), dst_path)
            scp.close()
            os.chown(dst_path, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
            os.chmod(dst_path, stat.S_IRWXU)

        self.runGuestSFTP(func, auth_info, src_path, dst_path)

    def removeFileInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        def func(transport, file_path):
            if self.isWindowsVM(vm_id, vm_user, vm_pwd, vm_address, vm_port):
                self.runGuestShell(auth_info, 'del "%s"' % toWinPath(file_path), 180 , True)
            else :
                sftp = paramiko.SFTPClient.from_transport(transport)
                sftp.unlink(file_path)

        self.runGuestSFTP(func, auth_info, file_path)

    def readEnvVariableInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, names=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        results = self.runGuestShell(auth_info, 'cmd.exe /c set')
        variables = []
        for result in results.split():
            try:
                key, value = result.split('=')
                if key == names:
                    variables.append(result)
            except:
                pass
        return variables

    def executeProgramInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, file_path=None, arguments=None, timeout=180):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        out_json = {}
        try:
            self.runGuestShell(auth_info, '%s %s' % (file_path, arguments), timeout)
        except GuestOSException as e:
            out_json['exit_code'] = e.errno
            out_json['exception'] = str(e)
        else:
            out_json['exit_code'] = 0

        return out_json

    def createDirectoryInVm(self, vm_id=None, vm_user=None, vm_pwd=None, vm_address=None, vm_port=None, dir_path=None):
        auth_info = {
            'vm_id': vm_id,
            'address': vm_address,
            'port': vm_port,
            'username': vm_user,
            'password': vm_pwd
        }

        auth_info = self._getAuthenInfo(vm_id, vm_user, vm_pwd, vm_address, vm_port)
        is_windows_vm = self.isWindowsVM(auth_info['vm_id'], auth_info['username'], auth_info['password'], auth_info['address'], auth_info['port'])
        vmio = VmIO(is_windows_vm, auth_info)
        return self.runGuest(vmio.createDirectory, auth_info, dir_path)

    def getMigrateCapability(self, host_id=None, host_obj=None):
        return {
            'vmotion': False,
            'vcenter': True
        }

    def getHostInterface(self, host_id=None):
        pass

    def queryIpFromDns(self, host_name):
        ips = self._run_cmd('[System.Net.Dns]::GetHostAddresses("%s")' % host_name)
        ips = toList(ips)
        if len(ips) == 0:
            raise Exception('Failed to query IP of [%s]' % host_name)
        return ips[0]['IPAddressToString']

    def runShell(self, command, timeout):
        return self._run_cmd(command, False)

    def runTaskGuestSFTP(self, task, f, auth_info, *args, **kwargs):
        try:
            return self.runGuestSFTP(f, auth_info, *args, **kwargs)
        except Exception as e:
            task.info.state = 'error'
            task.info.error = e
            raise e

    def runGuestSFTP(self, f, auth_info, *args, **kwargs):
        address, port, username, password = self._get_vm_connection_info(auth_info)

        try:
            transport = paramiko.Transport((address, port))
            transport.default_window_size = paramiko.common.MAX_WINDOW_SIZE
            transport.packetizer.REKEY_BYTES = pow(2, 40)
            transport.packetizer.REKEY_PACKETS = pow(2, 40)

            transport.connect(
                username = username,
                password = password)
            ret = f(transport, *args, **kwargs)
        except (IOError, socket.gaierror) as e:
            raise GuestOSException(e.errno, e.strerror, address=address)
        except SSHException as e:
            strerror = six.u(str(e))
            if 'Channel closed.' == strerror:
                strerror = 'Executing commands on a remote machine over SSH is not available'
            raise GuestOSException(errno.ECONNRESET, strerror, address=address)
        finally:
            try:
                transport.close()
            except:
                logger.error(traceback.format_exc())

        return ret

    def runGuest(self, f, auth_info, *args, **kwargs):
        address, port, username, password = self._get_vm_connection_info(auth_info)

        try:
            ret = f(*args, **kwargs)
        except (IOError, socket.gaierror) as e:
            raise GuestOSException(e.errno, e.strerror, address=address)
        except SSHException as e:
            strerror = six.u(str(e))
            if 'Channel closed.' == strerror:
                strerror = 'Executing commands on a remote machine over SSH is not available'
            raise GuestOSException(errno.ECONNRESET, strerror, address=address)

        return ret

    def runGuestShell(self, auth_info, command, timeout=180, chcp=False):
        address, port, username, password = self._get_vm_connection_info(auth_info)
        ssh = self.getSSHConnect(address, port, username, password)
        ret = runGuestShellBySSH(ssh, command, timeout, chcp)
        ssh.close()
        return ret

    def getSSHConnect(self, address, port, username, password):
        ssh = None
        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(
                address,
                port = port,
                username = username,
                password = password)
        except (IOError, socket.gaierror) as e:
            raise GuestOSException(e.errno, e.strerror, address=address)
        except SSHException as e:
            strerror = six.u(str(e))
            if 'Channel closed.' == strerror:
                strerror = 'Executing commands on a remote machine over SSH is not available'
            raise GuestOSException(errno.ECONNRESET, strerror, address=address)

        return ssh

    ####  util functions ###
    def _get_vm_connection_info(self, auth_info):
        return (
            auth_info['address'] if auth_info['address'] else self._get_vm_address(auth_info['vm_id']),
            auth_info['port'] if auth_info['port'] else 22,
            auth_info['username'],
            auth_info['password']
        )

    def _get_vm_address(self, vm_id):
        address = ''
        network = self._run_cmd(
            'Get-VM -Id %s '
            '| Select -ExpandProperty NetworkAdapters '
            '| Select IPAddresses' % vm_id)
        if 'IPAddresses' in network and 0 < len(network['IPAddresses']):
            address = network['IPAddresses'][0]
        else:
            raise GuestOSException(
                errno.EDESTADDRREQ,
                'Can not detect the address of guest OS. Please set it manually.')
        return address

    def _parse_datastore_data(self, datastore_id):
        datastore = self._run_cmd('Get-Volume -FilePath "%s\\"' % ntpath.normpath(datastore_id))
        if datastore == None:
            raise Exception('Cannot find datastore [%s]' % datastore_id)

        def escape(string):
            return string.replace('\\', '\\\\')

        obj = {}
        obj['name'] = self._run_cmd('Get-WmiObject Win32_Volume -Filter "DeviceID = \'%s\'"' % escape(datastore['UniqueId']))['Name']
        obj['accessible'] = True
        obj['capacity'] = datastore['Size']
        obj['free_space'] = datastore['SizeRemaining']
        obj['datastore_id'] = obj['id'] = datastore_id
        return obj

    def _parse_snapshot_data(self, vm_id=None, snap_id=None):
        if snap_id:
            cmd = 'Get-VMSnapshot -Id ' + snap_id
        else:
            cmd = 'Get-VM -Id ' + vm_id + '| Get-VMSnapshot'
        cmd += '| Select Id,Name,SnapshotType' \
            + ',@{N="CreateTime";E={Get-Date -Date $_.CreationTime -UFormat %s}}' \
            + ',@{N="VirtualDisk";E={,@($_.HardDrives|Select Path,DiskNumber,ControllerType,ControllerNumber,ControllerLocation)}}' \
            + ',@{N="ConfigFiles";E={,@(Get-ChildItem -LiteralPath "$($_.Path)\\Snapshots" -Filter ($_.Id.Guid + ".*") -File|Select FullName)}}'

        snapshots = []
        for snapshot in toList(self._run_cmd(cmd)):
            obj = {}
            obj['snap_id'] = snapshot['Id']
            obj['desc'] = SNAPSHOT_DESC
            obj['name'] = snapshot['Name']
            obj['type'] = snapshot['SnapshotType']
            obj['cbt_enable'] = True
            obj['create_time'] = snapshot['CreateTime']
            obj['virtual_disk'] = self._parse_virtual_disk(snapshot['VirtualDisk'])
            obj['config_files'] = self._parse_config_files(snapshot['ConfigFiles'])
            snapshots.append(obj)

        return snapshots[0] if snap_id else snapshots

    def _parse_vm_data(self, vm_id=None, get_detail=False, clustered=None):
        stableStatus = [
                2,      #OK
                3,      #Degraded
                5,      #PredictiveFailure
                10,     #Stopped
                11,     #InService
                15,     #Dormant
                32768,  #CreatingSnapshot
                32769,  #ApplyingSnapshot
                32770,  #DeletingSnapshot
                32771,  #WaitingToStart
                32772,  #MergingDisks
                32773,  #ExportingVirtualMachine
                32774   #MigratingVirtualMachine
                ]
        cmd = 'Get-VM'
        if vm_id:
            cmd += ' -Id ' + vm_id
        if clustered != None:
            cmd += '| Where IsClustered -eq ' + ('$true' if clustered else '$false')
        cmd += '| Select Id,Name,MemoryStartup,ProcessorCount,State,Status,PrimaryOperationalStatus,Version,IsClustered'
        cmd += ',@{N="ComputerName";E={hostname}}'
        if get_detail:
            cmd += ',@{N="NetworkAdapters";E={,@($_.NetworkAdapters|Select SwitchId,SwitchName)}},Generation' \
                +  ',@{N="VirtualDisk";E={,@($_.HardDrives|Select Path,DiskNumber,ControllerType,ControllerNumber,ControllerLocation)}}' \
                +  ',@{N="ConfigFiles";E={,@(Get-ChildItem -LiteralPath "$($_.Path)\\Virtual Machines" -Filter ($_.Id.Guid + ".*") -File|Select FullName)}}' \
                +  ',@{N="UsedSpace";E={($_,($_|Get-VMSnapshot)|Get-VMHardDiskDrive|Get-VHD|Measure-Object FileSize -Sum).Sum}}' \
                +  ',@{N="OsName";E={\n' \
                +      '$vm = Get-WmiObject -Namespace root\\virtualization\\v2 -Class Msvm_ComputerSystem | Where Name -eq $_.Id.Guid\n' \
                +      '$vm.GetRelated("Msvm_KvpExchangeComponent").GuestIntrinsicExchangeItems | % {\n' \
                +          '$node = ([XML]$_).SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Name\']")\n' \
                +          'if ($node.value -eq "OSName") {\n' \
                +              '$name = $node.SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Data\']").value\n' \
                +          '}\n' \
                +          'if ($node.value -eq "ProcessorArchitecture") {\n' \
                +              '$arch = $node.SelectSingleNode("/INSTANCE/PROPERTY[@NAME=\'Data\']").value\n' \
                +          '}\n' \
                +      '}\n' \
                +      'if ($name -and $arch) { $name + " (" + @("64", "32")[$arch -eq 0] + "-bit)" }\n' \
                +  '}}'

        vms = []
        for vm in toList(self._run_cmd(cmd, True, vm_id is not None)):
            try:
                obj = {}
                obj['type'] = 'vim.VirtualMachine'
                obj['name'] = vm['Name']
                obj['vm_id'] = obj['id'] = vm['Id']
                obj['os_name'] = vm['OsName'] if ('OsName' in vm and vm['OsName']) else 'Other'
                obj['is_template'] = False
                obj['vmtool_running'] = True
                obj['instance_uuid'] = vm['Id']
                obj['ram_mb'] = vm['MemoryStartup']/1024/1024
                obj['num_cpu'] = vm['ProcessorCount']
                obj['powered_on'] = vm['State'] == 2 # 2 = running, 3 = off
                obj['status'] = vm['Status']
                obj['status_ok'] = vm['PrimaryOperationalStatus'] in stableStatus
                obj['hw_version'] = vm['Version']
                obj['is_clustered'] = vm['IsClustered']
                obj['hypervisor_name'] = ''
                obj['is_vcenter'] = False

                if get_detail:
                    obj['folder_path'] = ''
                    obj['storage_size'] = vm['UsedSpace']
                    obj['virtual_disk'] = self._parse_virtual_disk(vm['VirtualDisk'], vm_id is None)
                    obj['config_files'] = self._parse_config_files(vm['ConfigFiles'])

                    obj['hypervisor'] = {
                        'host_id': vm['ComputerName'],
                        'name': vm['ComputerName'],
                        'standalone': True
                    }
                    obj['resourcepool'] = {
                        'is_vapp': False,
                        'name': vm['ComputerName'],
                        'resourcepool_id': vm['ComputerName']
                    }
                    if len(obj['config_files']) > 0:
                        folder_id = ntpath.dirname(ntpath.dirname(obj['config_files'][0]))
                        obj['folder'] = {
                            'name': ntpath.basename(folder_id),
                            'folder_id': toWinPath(folder_id),
                            'datacenter_id': ''
                        }
                    else:
                        obj['folder'] = {
                            'name': '',
                            'folder_id': '',
                            'datacenter_id': ''
                        }

                    obj['datastore'] = []
                    for disk in obj['virtual_disk']:
                        datastore_id = ntpath.dirname(disk['filename'])
                        obj['datastore'].append({
                            'name': ntpath.basename(datastore_id),
                            'datastore_id': toWinPath(datastore_id)
                        })
                    obj['network'] = []
                    for network in vm['NetworkAdapters']['value']:
                        obj['network'].append({
                            'name': network['SwitchName'],
                            'network_id': network['SwitchId']
                        })
                    obj['is_bios'] = vm['Generation'] == 1
                    obj['cbt_enabled'] = True

                vms.append(obj)
            except Exception:
                if vm_id:
                    raise
                else:
                    logger.error(traceback.format_exc())

        return vms[0] if vm_id else vms

    def _parse_config_files(self, data):
        if not data or 'value' not in data:
            return []

        files = []
        for file in data['value']:
            file = file['FullName'].lower()
            if file.endswith('vmcx') or file.endswith('vmrs') or file.endswith('vmgs'):
                files.append(toUnixPath(file))
        return files

    def _parse_virtual_disk(self, data, no_throw = False):
        if not data or 'value' not in data:
            return []

        disks = []
        for disk in data['value']:
            obj = {}
            if not disk['Path']:
                continue
            try:
                diskDirname = ntpath.dirname(disk['Path'])
                diskBasename = ntpath.basename(disk['Path'])

                if set(diskDirname) - set('[]') < set(diskDirname):
                    raise InvalidArgument('Path [%s] can not contain square brackets([]).' % disk['Path'])
                elif set(diskBasename) - set('[]\'') < set(diskBasename):
                    raise InvalidArgument('Path [%s] can not contain square brackets([]) or single quotes(\').' % disk['Path'])
                elif '\\' in disk['Path']:
                    vhd = self._run_cmd('Get-VHD -Path "' + escapePath(disk['Path']) + '" | Select Path,Size,@{N="Root";E={\n' \
                        +     '$path = $_.Path\n' \
                        +     'while ($path) {\n' \
                        +         '$disk = Get-VHD -Path $path\n' \
                        +         '$path = $disk.ParentPath\n' \
                        +     '}\n' \
                        +     '$disk|Select Path,VhdType,DiskIdentifier\n' \
                        + '}}')

                    vhd['Path'] = vhd['Path'].lower()
                    if not vhd['Root']:
                        raise VirtualDiskNotFound('Virtual disk root is not found.', id=disk['Path'])
                    vhd['Root']['Path'] = vhd['Root']['Path'].lower()
                    obj['key'] = vhd['Root']['DiskIdentifier']
                    obj['name'] = ntpath.basename(vhd['Root']['Path'])
                    obj['eager_scrub'] = True
                    obj['capacity'] = vhd['Size']
                    obj['filename'] = toUnixPath(vhd['Path'])
                    obj['filename_root'] = toUnixPath(vhd['Root']['Path'])
                    obj['thin_provision'] = vhd['Root']['VhdType'] == 3 # dynamic = 3, fixed = 2
                else:
                    vhd = self._run_cmd('Get-WmiObject Win32_DiskDrive | Where Index -eq ' + str(disk['DiskNumber']) + ' | Select Size')
                    obj['key'] = disk['Path'].lower()
                    obj['name'] = ''
                    obj['eager_scrub'] = True
                    obj['capacity'] = vhd['Size']
                    obj['filename'] = ''
                    obj['filename_root'] = disk['Path']
                    obj['thin_provision'] = False
                    obj['disk_mode'] = 'passthrough'
                obj['controller_type'] = 'IDE' if disk['ControllerType'] == 0 else 'SCSI'
                obj['bus_number'] = disk['ControllerNumber']
                obj['unit_number'] = disk['ControllerLocation']
                disks.append(obj)
            except:
                if not no_throw:
                    raise
                logger.error(traceback.format_exc())
                continue

        return disks

    def _run_cmd(self, cmd, to_json=True, err_stop=True, long_cmd=False):
        if to_json:
            cmd += '| ConvertTo-Json -Depth 10'

        errInfo = '$host.ui.WriteErrorLine(($_ | Select '\
                        'CategoryInfo,'\
                        'FullyQualifiedErrorId,'\
                        '@{N="Exception";E={,@($_.Exception | Select Message)}} |'\
                    'ConvertTo-Json))'
        if long_cmd:
            if not hasattr(self, 'mnt_src_path') or not hasattr(self, 'mnt_dst_path'):
                raise Exception("Please mount samba in hyperv_agent.cpp")

            cmd = '$ErrorActionPreference = "%s"\n' % ('Stop' if err_stop else 'Continue') \
                +     cmd

            sha1 = hashlib.sha1()
            sha1.update(six.ensure_binary(cmd))
            basename = sha1.hexdigest() + '.ps1'
            src_path = six.ensure_str(os.path.join(self.mnt_src_path, basename))
            dst_path = six.ensure_str(os.path.join(self.mnt_dst_path, basename))

            with codecs.open(src_path, 'w', 'utf-8') as file:
                file.write(u'\ufeff') # utf8 with BOM
                file.write(cmd)

            try:
                std_out = self._run_cmd('&"' + dst_path + '"', False, False, False)
                status_code = 0
            finally:
                try:
                    os.remove(src_path)
                except:
                    pass
        else:
            cmd = '$ErrorActionPreference = "%s"\n' % ('Stop' if err_stop else 'Continue') \
                + 'Try {\n' \
                +     cmd + '\n' \
                + '} Catch {\n' \
                +     '%s\n' % ('' if long_cmd else errInfo) \
                +     'exit 1\n' \
                + '}'

            enc_ps = b64encode(cmd.encode('utf_16_le')).decode('ascii')
            cmd_id = self.protocol.run_command(self.shell_id, 'powershell -encodedcommand %s' % enc_ps)
            std_out, std_err, status_code = self.protocol.get_command_output(self.shell_id, cmd_id)
            self.protocol.cleanup_command(self.shell_id, cmd_id)

        if status_code:
            if len(std_err):
                std_err = self._clean_error_msg(std_err)

            errmsg, err_id_list = self._get_msg_compare_vm(std_out)
            if len(errmsg):
                raise PowerShellException(
                    errmsg,
                    category_info='',
                    error_id=err_id_list)
            else:
                try:
                    except_err = getJsonFromString(std_err)
                    raise PowerShellException(
                        except_err['Exception']['value'][0]['Message'],
                        category_info=except_err['CategoryInfo'],
                        error_id=except_err['FullyQualifiedErrorId'])
                except ValueError:
                    raise PowerShellException("std_out: %s, std_err: %s, status_code: %d" % (std_out, std_err, status_code))

        if to_json and std_out:
            return json.loads(std_out)
        if std_out:
            return six.ensure_str(std_out.strip(), 'utf-8-sig', 'ignore')
        return None

    def _get_msg_compare_vm(self, msg):
        if not len(msg):
            return '', ''
        json_msg = json.loads(msg)

        class errReport:
            def __init__(self):
                self.msg = ''
                self.id_list = ''

        err = errReport()
        def append_err(item, err):
            if 'Message' in item:
                if len(err.msg):
                    err.msg += ' '
                err.msg += item['Message']
            if 'MessageId' in item:
                if len(err.id_list):
                    err.id_list += ','
                err.id_list += str(item['MessageId'])

        if isinstance(json_msg, dict):
            append_err(json_msg, err)
        elif isinstance(json_msg, list):
            for item in json_msg:
                append_err(item, err)

        return err.msg, err.id_list

    def _clean_error_msg(self, msg):
        """converts a Powershell CLIXML message to a more human readable string
        """
        if msg.startswith(b"#< CLIXML\r\n"):
            msg_xml = msg[11:]
            try:
                msg_xml = self._strip_namespace(msg_xml)
                root = ET.fromstring(msg_xml)
                nodes = root.findall("./S")
                new_msg = ""
                for s in nodes:
                    new_msg += s.text.replace("_x000D__x000A_", "\n")
            except Exception as e:
                print("Warning: there was a problem converting the Powershell"
                      " error message: %s" % (e))
            else:
                if len(new_msg):
                    msg = new_msg.strip()
        return msg

    def _strip_namespace(self, xml):
        """strips any namespaces from an xml string"""
        try:
            p = re.compile(b"xmlns=*[\"\"][^\"\"]*[\"\"]")
            allmatches = p.finditer(xml)
            for match in allmatches:
                xml = xml.replace(match.group(), b"")
            return xml
        except Exception as e:
            raise Exception(e)

    def _get_task(self, task_id):
        return self.task_pool[task_id]

    def _put_task(self, task):
        self.task_pool[task.info.key] = task
        return task.info.key

    def _del_task(self, task_id):
        del self.task_pool[task_id]
