#-*-coding: utf-8 -*-
import os, fcntl
import json, time, socket
import logging
import logging.handlers
import errno
from enum import Enum
from .defines import *

logger = logging.getLogger(VM_LOGGER_NAME)

class IpcClient:
    def __init__(self):
        self.sock = None
        self.REQUEST_END_MAGIC = "^^2552*1814^^"  # like REQUEST_END_MAGIC in c++ API

    def __del__(self):
        self.sock.close()

    def connect(self, sock_path):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(sock_path)

    def send(self, json_req):
        str_req = json.dumps(json_req)
        str_req += self.REQUEST_END_MAGIC
        self._send(str_req)
        return self._recv()

#-------------- Private Function --------------
    def _recv(self):
        msg = ""
        while True:
            try:
                data = self.sock.recv(1024).decode()
                if not data:
                    break
                msg += data
                if msg.find(self.REQUEST_END_MAGIC):
                    break
            except (socket.error) as e:
                if e.errno != errno.EINTR:
                    logger.error(e)
                    raise
        return msg

    def _send(self, data):
        data = data.encode()
        bytes_to_write = len(data)
        sent = 0
        while sent < bytes_to_write:
            try:
                chunk_sent = self.sock.send(data)
                sent += chunk_sent
                data = data[chunk_sent:]
            except (socket.error) as e:
                if e.errno != errno.EINTR:
                    logger.error(e)
                    raise


class LogType(Enum):
    INTERNAL_ERROR = 1000
    VM_RESTORE_GUEST_NOT_ENOUGH_SPACE = 3061
    REASON_GUEST_OS = 3500
    PATH_RESTORE = 6003
    PATH_RESTORE_ERROR = 6006
    PATH_RESTORE_NOT_SUPPORT = 6011
    PATH_RESTORE_META_ERROR = 6014

class LogLevel(Enum):
	ERROR          = 0
	WARNING        = 1
	INFO           = 2
	DEBUG          = 3


class ActivityLog:
    def __init__(self, result_id):
        self.result_id = result_id
        self.logLevel = {
            LogType.PATH_RESTORE: LogLevel.INFO,
            LogType.PATH_RESTORE_ERROR: LogLevel.ERROR,
            LogType.PATH_RESTORE_META_ERROR: LogLevel.WARNING,
            LogType.VM_RESTORE_GUEST_NOT_ENOUGH_SPACE: LogLevel.ERROR,
            LogType.PATH_RESTORE_NOT_SUPPORT: LogLevel.WARNING,
            LogType.REASON_GUEST_OS: LogLevel.ERROR
        }

    def add_result_detail(self, log_type, other_params = {}):
        if 0 == self.result_id:
            return

        self.ipc = IpcClient()
        self.ipc.connect("/run/synoabk/synoabk-log.sock")
        log = {}
        log["result_id"] = self.result_id
        log["result_detail_id"] = 0
        log["log_level"] = self.logLevel[log_type].value
        log["log_type"] = log_type.value
        log["log_time"] = time.time()
        log["other_params"] = other_params
        log["error_code"] = 0
        request = {}
        request["action"] = "add_activity_result_detail"
        request["log"] = log
        response = self.ipc.send(request)

